unit Inference;


interface

uses
  SysUtils, Windows, Messages, Classes, Graphics, Controls, Forms, Dialogs;


const

cforall = chr(247);    { Characters constants of special logic }
cexists = chr(246);    { and infix operators.                  }
cOrdand = chr(217);
cOrdor = chr(218);
cimp = chr(244);
ceqv = chr(243);
cneqv = chr(225);
cleqv = chr(223);
cgeqv = chr(221);
cneg = chr(172);
csup = chr(228);
cinf = chr(231);
cand = chr(253);
cor = chr(252);
cDegSep = ':';
cFmlSep = ';';

type

ErrorType = (OK, illchar, missoper, missexp, missbra, unexpbra, endcomp,
                insfail, mem, missdelim, wrongmol, InvMeth, InvRealVal,
                InvIntVal, InvIDTerm, InvVar, WrongDegExp);

StrategyType = (SrtLinear, SrtModifLinear, SrtBreadth);

{ Compile error codes, see XComp proc. }


TDoProc =
  function(o : TObject; var tp : char; argums : array of const) : TObject;
     { Template for functions used as executive procedures
       for the recursive browsing of the syntactical tree.(ST) }

{ Logical quantifier object (contains variables). }

TQuant = class(TList)
  procedure RPrint(var st : string);
    { RPrint recursively prints the quantifier and containing variables }
  procedure RFree; virtual;
    { RFree Recursively frees the q. and all the dependencies }
  function Dive(tp : char; proc : TDoProc; argums : array of const) : TQuant;
    { Dive serves for several types of the passes thru the ST. }
  procedure Swap;
    { Swap is required for self resolution during simplification of ST}
end;

TSubClass = class of TSub;

{ General class of logical connective. }

TSub = class
    Neg : boolean;  { Flag of logical negation. }
    Ev : shortint;  { Indicator of the subformula logical value. }
    TempEv : shortint; {Temporal evaluation - used for self resolution}
    Pol : shortint; { Stores the polarity of the node. }
    L : TSub;
    R : TSub;         { Left and right subtree. }
    Ac : TObject;    { Parent object (logical connective). }
    Q : TQuant;     { Quantifier containing variables at this level. }
    procedure RFree; virtual;
    function GetSymbol : string; virtual; abstract;
      { Returns relevant character. }
    function GetP : integer; virtual; abstract;
      { Returns connective priority. }
    procedure RPrint(var st : string); virtual;
    procedure Negate; { Logically negates the node. }
    procedure AppendToQuant(var Q1 : TQuant);
      { Appends variables to another node. }
    function Dive(tp : char; proc : TDoProc; argums : array of const) : TSub;
      virtual;
    function SubstTo : TSub; { Substitutes a subtree with its superior node. }
    procedure EvalPol; { Evaluates polarity. }
    procedure OptQ;
end;

{ Logical conjunction node. }
TCon = class(TSub)
   function GetSymbol : string; override;
   function GetP : integer; override;
end;

{ Logical disjunction node. }
TDis = class(TSub)
   function GetSymbol : string; override;
   function GetP : integer; override;
end;

{ Logical implication node. }
TImp = class(TSub)
   function GetSymbol : string; override;
   function GetP : integer; override;
end;

{ Logical equivalence node. }

TVariable = class;

{ Base class for all data types data and identifiers. }
TId = class
  procedure RPrint(var st : string); virtual; abstract;
  procedure RFree; virtual;
  function Compare(X : TId) : Integer; virtual; abstract;
  function Dive(tp : char; proc : TDoProc; argums : array of const) : TId;
  function OccurCheck(X : TVariable) : boolean;
    { Checks for an occurence of a variable in the id. }
  function UnifyId(X : TId) : integer;
    { Unifies two identifiers. }
end;


{ Abstract class of a number. }
TNumber = class(TId)
  function GetNumber : extended; virtual; abstract;
end;


TReal = class(TNumber)
  Id : Extended;
  constructor Create(i : string);
  constructor CreateN(i : extended);
  procedure RPrint(var st : string); override;
  function Compare(X : TId) : Integer; override;
  function GetNumber : extended; override;
end;

TInteger = class(TNumber)
  Id : Longint;
  constructor Create(i : string);
  constructor CreateN(i : longint);
  procedure RPrint(var st : string); override;
  function Compare(X : TId) : Integer; override;
  function GetNumber : extended; override;
end;

{ Functor base class. }
TFunc = class(TId)
  function Compare(X : TId) : Integer; override;
end;

{ Functor with arguments. }
TFunctor = class(TFunc)
  Functor : string; { Functor id. }
  Params : TList;   { Arguments. }
  constructor Create(i : string);
  procedure RPrint(var st : string); override;
  procedure RFree; override;
  destructor Destroy; override;
end;

{ Functor without arguments. }
TFunctorA = class(TFunc)
  Functor : string;
  constructor Create(i : string);
  procedure RPrint(var st : string); override;
end;

{ Unoptimized variable class. }
TVar = class(TId)
  Id : string; { Name of the variable. }
  constructor Create(i : string);
  procedure RPrint(var st : string); override;
  function Compare(X : TId) : Integer; override;
end;

{ Optimized variable class - attached to its occurences in expressions. }
TVariable = class(TId)
  Id : TId; { Substitution of the variable by an expression. }
  Ex : char;  { Indicates quantification of a variable. }
  New : TVariable; { Temporal pointer used in copying of the variable. }
  NewId : TId; { Temporal pointer determining substitution. }
  Used : boolean; { Flag of variable usage in expressions. }
  UnSub : boolean; { Flag determining substitution. }
  Ap : Boolean; { Used during unification. }
  Master : TSub; { Master subformula of the variable. }
  Watch : TVariable; { Link to goal requested source variable. }
  SubId : integer;
  Source : boolean; { Variable is from source formula}
  IsInMGU : boolean; {Variable is present in current MGU}
  identifier : longint; {Unique identifier of every source variable /unification}
  Name : string;
  constructor Create(i : TVar);
  constructor CreateV(i : TVariable);
  constructor CreateS(i : string);
  procedure RPrint(var st : string); override;
  procedure PrintAs(var st : string); { Prints the variable substitution. }
  destructor Destroy; override;
  function Compare(X : TId) : Integer; override;
  function CompInMgu : TId; { Checks for current substitutions. }
  function MAssigned : boolean;
    { Checks for the assignment of all the superior variables. }
  procedure RFree; override;
end;

  { Atomic formula class. }
TAtom = class(TSub)
      Id : TId; { Identifier of a formula. }
      procedure RPrint(var st : string); override;
      procedure RFree; override;
      function GetP : integer; override;
      function Dive(tp : char; proc : TDoProc; argums : array of const) : TSub;
        override;
        { Unification of two atoms. }
      function UnifyS(X : TAtom) : integer;
end;

TPLProgram = class;

  { Predicate logic formula class. }
TFALFormula = class
   Cont : TSub; { Contents of the formula. }
   Symbolic : string;
   Owner : TPLProgram; {Master TPLProgram of the formula}
   Parent1 : TFALFormula;
   Parent2 : TFALFormula;
     { Parents of the formula in the meaning of resolution rule. }
   MF, ML : TAtom; { First and last atom. }
   Degree : extended; {Truth valuation in Lukasiewicz algebra}
   Support : boolean;  { Indicator for the Supprot Set Strategy. }
   DisabledDCF : boolean; {Flag determining formula killed by DCF}
   constructor Create(ow : TPLProgram);
   destructor Destroy; override;
   procedure Print(var s1 : string);
   procedure PrintRefutation(var FN1 : TFALFormula); {Prints current refutation degree}
   procedure PostComp; { Some transformations needed after compilation. }
   function
     Dive(tp : char; proc : TDoProc; argums : array of const) : TFALFormula;
   function
     DiveD(tp : char; proc : TDoProc; argums : array of const) : TFALFormula;
     { Same as dive, but it walks only by atoms and inferior nodes. }
   function Resolve(X : TFALFormula; T1, T2 : TAtom; premise : boolean) : TFALFormula;
     { Resolves two formulas with given unification. }
   procedure CompleteUnify(var T1:TAtom); {Complete unification / only self resolution!}
   function ResolveAWC(X : TFALFormula; T1 : TAtom; premise : boolean) : TFALFormula; {Self resolution returning the same formula as resolvent}
   function ResolveAWCFirst(X : TFALFormula; T1 : TAtom; premise : boolean) : TFALFormula;
   {Self resolution returning new formula as a resolvent}
   function OptimizeX : TFALFormula;
     { Optimization leading to evaluation of the formula. }
   procedure AllPosRes;
   procedure AllPosRes2;
   procedure AllPosRes3;
   procedure AllPosRes4;
     { Several types of resolution strategies. }
   function InsertToR : TFALFormula;
     { Checks for redundancy of the formula in the set of resolvents. / DCF}
   function InsertToREqv : TFALFormula;
     { Checks for equivalence of the formula in the set of resolvents. / DCF}
   function TryToRes(X : TFALFormula) : integer;
   function TryToRes2(X : TFALFormula) : integer;
   function TryToRes3(X : TFALFormula) : integer;
     { Several types of resolving two formulas. }

   procedure DCFKill;
   {Performs DCF Kill check /consequent former resolvent is disabled}
   procedure DCFKillEqv;
   procedure MakeSymbolic;
   function TrivialCheck(X : TFALFormula):boolean;

   function IsValidPair(X : TFALFormula) : boolean;
     { Checks for validity of two formulas in the meaning of Filtration st. }
   function CreateVF(X : TFALFormula; T : TSubClass) : TFALFormula;
     { Creates virtual formula for redundancy checking. }
   procedure DestroyVF(X1, X2 : TFALFormula);
     { Destroys and repairs original formulas from Virtual Formula}
   procedure Simplify;
     { Simplifies the formula in several ways. }
   procedure PrintVars;
     { Prints requested variables of the formula. }
   function PrintVarsToStr:string;
     { Prints requested variables of the formula. }
   procedure Add(X:TFALFormula;T:TSubClass);
end;

  { Special case of a formula - query. }
TQuery = class(TFALFormula)
   procedure EvalQuery; { Causes evaluation of the query. }
   procedure NegQ; { Negates query. }
end;


  { Predicate logic program. }
TPLProgram = class

  ListF : TList;  { List of source formulas and queries. }
  ListR : TList;  { List of resolvents. }
  MGUn1  : TList;  { Temporal store for unification results. }
  Err : ErrorType; { Temporal store for compilation error. }
  Localpos : Longint; { Position of an error. }
  Strategy : StrategyType; { Type of the resolution strategy. }
  {properties}
  InputErrorPos : longint;
  OutPutLines : ^TStrings;
  InputContent : PChar;
  DCF : integer;
  StopF : boolean;
  Proved : boolean;

  Interactive : boolean;
  Time : boolean;
  Memory1 : boolean;
  Axioms : boolean;
  Resolvents : boolean;
  Stat : boolean;
  Checkby1 : boolean;
  Checkby2 : boolean;
  Without1 : boolean;
  Support1 : boolean;
  Filtra1 : boolean;
  Unsim : boolean;
  Progress : boolean;
  Sources: boolean;
  ExitOn1 : boolean;
  ExitOn3 : boolean;
  SrCheck1 : boolean;
  General : boolean;
  RedCheck1 :boolean;
  OneRef1 : boolean;
  Quant1 : boolean;

  constructor Create;
  destructor Destroy; override;
  procedure Generate; { Generates set of compiled formulas. }
  procedure XComp(var infix : PChar; var F : TFALFormula);
    { Compilation core. }
  procedure ClearFormulas;
  procedure ClearResolvents;
  procedure Consist; { Encapsulates consistency of the ListF checking. }
end;

var CurProg : TPLProgram;  { Current program link. }

var mgu1 : TLIst; NewQ : TQuant; BestDegree : extended;
    BestSubst : string;
var infer : boolean; rsv1, rsv2 : string;
    rsvnum, ignnum, numsub, numsim, nummol, numKill, varident,
    numfl, lastr1, lastr2, numsimX : longint; FX : TFALFormula;
    retf, slu, und : boolean;


implementation

const
  numeric = ['0'..'9'];
  lower = ['a'..'z'];
  upper = ['A'..'Z', '_'];
  alnum = numeric+lower+upper;
  ignore = [#1..#32];
  InfixPred = ['=', '<', '>', cneqv, cleqv, cgeqv];
  InfixOper = ['+', '-', '*', '/'];

type
     PSStackNode = ^TSStackNode; { Stack node containing tree node }
     TSStackNode = record        { during compilation. }
      Node : TObject;            { Pointer to tree node. }
      Next : PSStackNode;        { Pointer to next stack node. }
     end;
     PSStack = PSStackNode;

var ch : char;                   { Current character. }
    cpos, inum, dontcare : longint;
    parse : PChar; ident : string;
    LastMol, FirstMol : TAtom;   { Last and First atom. }
    Sub : TSub; MaxMem,MinMem : longint;
    VPSStack : PSStack; pom, pom1, pom2, garbage : PSStackNode;
    Error : ErrorType; query : boolean;
    opt : boolean;   { Indicates succesful optimizitation. }


function UsedMem : longint;
var x : THeapStatus;
begin
  x := GetHeapStatus;
  UsedMem := x.TotalAllocated;
end;

procedure StackInit;
begin
  VPSStack := nil;
end;

procedure Put(N : TObject);
begin
  new(pom2);
  pom2^.next := VPSStack;
  VPSStack := pom2;
  VPSStack^.node := N;
end;

function Cut : pointer;
begin
  if VPSStack <> nil then
  begin
    pom1 := VPSStack;
    VPSStack := VPSStack^.next;
    Cut := pom1;
  end;
end;

procedure SInsert(p : PSStackNode);  { Inserts existing stack node. }
begin
  p^.next := VPSStack;
  VPSStack := p;
end;

procedure AssignSub(Ob : TObject);   { Assigns formula its subtrees. }
begin
  Put(Ob);
  pom := Cut;
  TSub(pom^.node).R := TSub(VPSStack^.node);
  TSub(VPSStack^.node).Ac := pom^.node;
  garbage := Cut;
  dispose(garbage);
  TSub(pom^.node).L := TSub(VPSStack^.node);
  TSub(VPSStack^.node).Ac := pom^.node;
  garbage := Cut;
  dispose(garbage);
  SInsert(pom);
end;

procedure AssignQToQ(Ob : TQuant; c : char); { Assigns quant. its variables.}
begin
  Ob.Add(TVariable.Create(TVar(VPSStack^.node)));
  TVariable(Ob.Items[Ob.count-1]).Ex := c;
  TVariable(Ob.Items[Ob.count-1]).Source := true;
  TVar(VPSStack^.node).Free;
  garbage := Cut;
  dispose(garbage);
end;

procedure AssignQToSub;
begin
  garbage := VPSStack^.next; VPSStack^.next := VPSStack^.next^.next;
  TSub(VPSStack^.node).AppendToQuant(TQuant(garbage^.node));
  dispose(garbage);
end;

procedure AssignTtoA(Ob : TAtom);
begin
  Ob.Id := TId(VPSStack^.node);
  garbage := Cut;
  dispose(garbage);
end;

procedure AssignAtom(Ob : TAtom);
begin
  if LastMol <> nil then
  begin
    Ob.L := LastMol;
    LastMol.R := Ob;
  end
  else FirstMol := Ob;
  Put(Ob); LastMol := Ob;
end;

procedure AssignTerms(Ob : TObject; NT : integer);
  { Assigns arguments to a term. }
var px : integer;
begin
  Put(Ob); px := 0;
  pom := Cut;
  while px < NT do
  begin
    TFunctor(pom^.node).Params.Insert(0, VPSStack^.node);
    garbage := Cut;
    dispose(garbage); inc(px);
  end;
  SInsert(pom);
end;

procedure NegateF; { Negates a formula. }
var i:integer;
begin
  if (VPSStack <> nil)and(VPSStack^.node is TSub) then
  begin
    TSub(VPSStack^.node).neg := not(TSub(VPSStack^.node).neg);
    if TSub(VPSStack^.node).Q<>nil then
    begin
      for i := 0 to TSub(VPSStack^.node).Q.Count-1 do
        if TSub(VPSStack^.node).Q.Items[i] <> nil then
          case TVariable(TSub(VPSStack^.node).Q.Items[i]).Ex of
            cexists : TVariable(TSub(VPSStack^.node).Q.Items[i]).Ex := cforall;
            cforall : TVariable(TSub(VPSStack^.node).Q.Items[i]).Ex := cexists;
          end;
    end;
  end;
end;

procedure GetCharI;  { Reads a character.}
 begin
   if Error = OK then
     begin
       inum := 0;
          ch := parse[cpos];
             if ch <> #0 then begin inc(cpos); inc(inum); end;
             if ((ch = cFmlSep)) and (parse[cpos] in ignore + [#0]) then
               ch := #0;
     end;
 end;

procedure GetChar;  { Reads a character and omits blank characters. }
 begin
   if Error = OK then
     begin
       inum := 0;
       repeat
            ch := parse[cpos];
            if ch <> #0 then begin inc(cpos); inc(inum); end;
            if ((ch = cFmlSep)) and (parse[cpos] in ignore + [#0]) then
              ch := #0;
       until not(ch in Ignore);
     end;
 end;


procedure IDTerm; forward;
procedure IDBase; forward;


procedure Number;
var point : boolean;
begin
  point := false;
  while ch in numeric do
  begin
    ident := ident+ch; GetCharI;
    if ch in ['e', 'E', DecimalSeparator] then
    begin
      point := true;
      ident := ident+ch; GetCharI;
      if ch in ['+', '-'] then
      begin
        ident := ident+ch; GetCharI;
      end;
    end;
  end;
  if ch in ignore then GetChar;
  if point then
     AssignTerms(TReal.Create(ident), 0)
  else AssignTerms(TInteger.Create(ident), 0);
  ident := '';
end;

procedure Variable;
begin
  while ch in alnum do
  begin
    ident := ident+ch;
    GetCharI;
  end;
  if ch in ignore then GetChar;
  if ident = '_' then
    begin
     AssignTerms(TVar.Create(ident+'#'+IntToStr(dontcare)), 0);
     inc(dontcare);
    end
  else AssignTerms(TVar.Create(ident), 0);
  ident := '';
end;

procedure Functor;
var L1 : TId; N1 : integer;
begin
  while (ch in alnum) do
    begin
      ident := ident+ch;
      GetCharI;
    end;
  N1 := 0;
  if (ch = '(') then
    begin
      GetChar;
      if (ch <> ')') then
      begin
        L1 := TFunctor.Create(ident); ident := '';
        IDTerm; inc(N1);
        while ch = ',' do
        begin
          GetChar; IDTerm; inc(N1);
        end;
        if (ch <> ')')and(error = OK) then error := missbra;
      end
      else if (ident = '(') then
      begin
        L1 := TFunctorA.Create(ident); ident := '';
      end
      else
      begin
        Error := InvIdTerm; Exit;
      end;
      GetChar;
    end
    else
    begin
      if Error = OK then L1 := TFunctorA.Create(ident);
      ident := '';
    end;
  if Error = OK then AssignTerms(L1, N1) else
    if L1 <> nil then L1.Free;
  if ch in ignore then GetChar;
end;

procedure IDTerm2;
var L1 : TId;
begin
  if Error = OK then
  begin
    IDBase;
     while ((ch = '*')or(ch = '/'))and(error = OK) do
       begin
         L1 := TFunctor.Create(ch);
         case ch of
         '*' : begin
                Getchar;
                IDBase; 
              end;
         '/' : begin
                Getchar;
                IDBase;
              end; 
         end; 
         if error = OK then AssignTerms(L1, 2) else L1.Free;
       end;
  end; 
end;

procedure IDTerm; 
var L1 : TId;
begin
  if Error = OK then
  begin
    IDTerm2; 
     while ((ch = '+')or(ch = '-'))and(error = OK) do
       begin
         L1 := TFunctor.Create(ch);
         case ch of
         '+' : begin
                Getchar;
                IDTerm2; 
              end; 
         '-' : begin
                Getchar; 
                IDTerm2; 
              end;
         end;
         if error = OK then AssignTerms(L1, 2) else L1.Free;
       end; 
  end;
end;

procedure IDBase;
begin
  if error = OK then
  begin
    case ch of
      'A'..'Z', '_' : Variable;
      'a'..'z' : Functor; 
      '0'..'9' : Number; 
      '+' :
      begin
        GetChar; 
        IDBase; 
        if Error = OK then
        if (VPSStack.node is TReal) then
        TReal(VPSStack.node).Id := TReal(VPSStack.node).Id
        else if (VPSStack.node is TInteger) then
        TInteger(VPSStack.node).Id := TInteger(VPSStack.node).Id
        else AssignTerms(TFunctor.Create('+'), 1);
      end;
      '-' : 
      begin
        GetChar;
        IDBase;
        if Error = OK then
        if (VPSStack.node is TReal) then
        TReal(VPSStack.node).Id := -TReal(VPSStack.node).Id
        else if (VPSStack.node is TInteger) then
        TInteger(VPSStack.node).Id := -TInteger(VPSStack.node).Id
        else AssignTerms(TFunctor.Create('-'), 1);
      end;
      '(' : 
      begin
        GetChar; 
        IDTerm;
        if (ch <> ')')and(Error = OK) then Error := MissBra;
        GetChar;
      end;
    else
      if (error = OK) then Error := InvIDTerm;
    end;
  end;
end; 


procedure InfixPr;
var ch1 : char;
begin
  if Error = OK then
  begin
    ch1 := ch; GetChar;
    IDTerm; 
    if error = OK then AssignTerms(TFunctor.Create(ch1), 2); 
  end;
end; 

procedure Atom;
var LF : TAtom;
begin
  if error = OK then
 begin
   IDTerm;
   if Error = OK then
   begin
     LF := TAtom.Create;
   end
   else Exit;
   begin
     if ch in ignore then GetChar;
     if ch in InfixPred then
     begin
       InfixPr;
       if error = OK then
       begin
         AssignTToA(LF); AssignAtom(LF);
       end
       else
       begin
         LF.RFree; Exit;
       end;
     end
     else
     if (VPSStack <> nil)and((VPSStack^.node) is TFunc) then
     begin
       if error = OK then
       begin
         AssignTtoA(LF);
         AssignAtom(LF);
       end
       else
       begin
         LF.RFree; Exit;
       end;
     end
     else
     begin
       if error = OK then
       begin
         error := wrongmol;
         LF.RFree; 
       end; 
     end; 
   end; 
 end; 
end;

procedure Quant;
var O1 : TQuant; ch1 : char;
begin
  if Error = OK then
  begin
    O1 := TQuant.Create;
    while ch in [cforall, cexists] do
    begin
      ch1 := ch;
      GetChar;
      if ch in Upper then Variable else Error := InvVar;
      if Error = OK then AssignQtoQ(O1, ch1);
      while ch = ',' do
      begin
        GetChar;
        if ch in Upper then Variable else Error := InvVar;
        if Error = OK then AssignQtoQ(O1, ch1);
      end;
    end;
    if Error = OK then Put(O1) else O1.RFree;
  end;
end; 

procedure Imp;forward;

procedure ICE;
begin
if error = OK then
 begin
  case ch of
   cneg : begin
         Getchar; 
         ICE;
         NegateF;
       end;
   cforall, cexists :
        begin
          Quant;
          ICE;
          if Error = OK then AssignQtoSub; 
        end;
   'a'..'z', 'A'..'Z', '_', '0'..'9', '(', '+', '-' :
       begin
         Atom;
       end;
    '[' : begin
         GetChar;
         Imp; 
         if (ch <> ']')and(error = OK) then Error := missbra;
         Getchar; 
       end
   else if error = OK then Error := missexp; 
   end;
 end;
end;

procedure Con; 
begin
 if Error = OK then
   begin
     ICE; 
     while (ch = cand)and(error = OK) do
       begin
        Getchar;
        ICE;
        if error = OK then AssignSub(TCon.Create);
       end;
   end;
end; 

procedure Dis; 
begin
 if Error = OK then
   begin
     Con;
     while (ch = cor)and(error = OK) do
       begin
        Getchar; 
        Con;
        if error = OK then AssignSub(TDis.Create);
       end; 
   end;
end; 

procedure Imp;
begin
 if Error = OK then
   begin
     Dis;
     while (ch = cimp)and(error = OK) do
       begin
        Getchar; 
        Dis; 
        if error = OK then AssignSub(TImp.Create);
       end;
   end;
end;


procedure DestroyNode(x : TObject);
begin
  if Assigned(x) then
  if x is TSub then
  TSub(x).RFree else
  if x is TVariable then
  TVariable(x).Free else  { Exceptional behavior. }
  if x is TId then
  TId(x).RFree else
  if x is TQuant then
  TQuant(x).RFree;
end;

procedure DisposeStack;
var pte : PSStackNode;
begin
  if VPSStack <> nil then
   begin
     pte := Cut;
     dispose(pte);
   end;
  while VPSStack <> nil do
   begin
    pte := Cut;
    DestroyNode(pte^.node);
    dispose(pte);
   end;
end;

function Max(xx1, xx2 : shortint) : shortint;
begin
  if xx1 > xx2 then Result := xx1 else Result := xx2;
end;

function Min(xx1, xx2 : shortint) : shortint;
begin
  if xx1 < xx2 then Result := xx1 else Result := xx2;
end;

function LukMult(x1, x2 : extended) : extended;
begin
  Result := x1 + x2 - 1;
  if Result < 0 then Result := 0;
end;

var deg : extended; DegStr, SepPos : PChar;
    time1, time2 : double;

 procedure TPLProgram.XComp;
 begin
  F := nil; ident := ''; Sub := nil;
  LastMol := nil; FirstMol := nil; query := false;
  Deg := 1;
  parse := infix;
  StackInit;
  cpos := 0; dontcare := 0;
  Error := OK;
  Getchar;
  if (ch = '?')and(infix[cpos] = '-') then
  begin
   query := true;
   GetCharI; GetChar;
  end;
  if Error = OK then Imp;  { Go to parser / generator. }
  if (ch = ')')and(error = OK) then Error := unexpbra;

  //may contain formula degree
  if (ch = cDegSep)and(error = OK) then
  begin
    try
      SepPos := StrPos(@(parse[cpos]),PChar(string(cFmlSep)));
      DegStr := StrAlloc(SepPos - @(parse[cpos]) + 1);
      StrLCopy(DegStr, @(parse[cpos]), SepPos - @(parse[cpos]));
      Deg := StrToFloat(DegStr);
      if (Deg < 0) or (Deg > 1) then error := wrongDegExp;
      StrDispose(DegStr);
      cpos := cpos + SepPos - @(parse[cpos]) + 1;
      ch := #0;
    except
      error := wrongDegExp;
      StrDispose(DegStr);
    end;
  end;

  if (ch = #0)and(error = OK)and(parse[cpos-1] = cFmlSep) then
     Error := endcomp;
  if (ch = #0)and(error = OK)and(parse[cpos-1] <> cFmlSep) then
     Error := missdelim;
  if (ch <> #0)and(error = OK) then Error := missoper;
  if Error = endcomp then
   begin
      Error := OK;
      sub := TSub(VPSStack^.Node);
      dispose(VPSStack);
      if query then F := TQuery.Create(self)
      else F := TFALFormula.Create(self);
      F.Cont := sub;
      sub.Ac := F;
      F.MF := FirstMol;
      F.ML := LastMol;
      F.Degree := Deg;
      F.PostComp;                 { Make the formula. }
      GetChar;
      if ch <> #0 then LocalPos := cpos-1 else LocalPos := cpos;
   end
  else
     begin
       if VPSStack <> nil then
         begin
           DestroyNode(VPSStack^.node);
         end;
      DisposeStack;
      LocalPos := cpos-inum-1;
      end;
 if error <> OK then time2:= Now;
 err := error;
end;

constructor TPLProgram.Create;
begin
  DCF := 10;
  StopF := false;
  ListF := TList.Create;
  ListR := TList.Create;
  MGUn1 := TList.Create; mgu1 := mgun1;
end;


procedure TPLProgram.Generate;
var p1, p2 : longint; temp : PChar; Form, Form2 : TFALFormula;
    infstr : string;
begin
  p1 := 0; p2 := 0; temp := InputContent;
  InputErrorPos := 0;
  CurProg := self;
  ClearFormulas;
  MinMem := UsedMem; MaxMem := MinMem;
  time1 := Now; time2 := 0;
  while temp[0] <> #0  do         { Compiles until end is founded. }
  begin
    Application.ProcessMessages;
    if stopf then Exit;
    Form := nil;
    XComp(temp, Form);
    if Form <> nil then
    begin

      Form.Simplify;
      if Form is TQuery then
      begin
        TQuery(Form).EvalQuery;   { Evals if it is query. }
      end
        else ListF.Add(Form);     { Or adds to ListF. }
      p1 := LocalPos;
      temp := @(temp[p1]);
      p2 := p2+p1;
    end
    else begin
      InputErrorPos := p2+LocalPos+1;
      ClearFormulas; temp := PChar('');
    end;
  end;
  if time2 = 0 then time2 := Now;

  if Proved then
  begin
    infstr := 'Best refutation degree: '+
           FloatToStr(BestDegree);
    if BestSubst <> '' then infstr := infstr+'('+BestSubst+')';
    OutputLines.Add(infstr);
  end;

  if Time then
    infstr := 'Solving time : '+
           FloatToStrF((time2-time1)*84600, ffFixed, 7, 2)+' s. ';
  if Memory1 then
  begin
    infstr := infstr+'Used memory for proof: ' +IntToStr(MaxMem-MinMem)+' B.';
    OutputLines.Add(infstr);
    infstr := 'Total allocated for GERDS: ' +IntToStr(UsedMem)+' B.';
    OutputLines.Add(infstr);
  end;

end;


procedure TPLProgram.Consist;
var i, j : longint; XF : TFALFormula; XM : TAtom; s : string;
begin
  ClearResolvents;
  BestDegree := 0; BestSubst := '';
  infer := true; rsvnum := 0; lastr1 := 0;
  lastr2 := 0;
  ignnum := 0; rsv1 := ''; numsub := 0; nummol := 0; numKill := 0; numsim := 0;
  numsimX := 0; mgu1.clear; slu := false; und := false; varident := 0;
  retf := false; Proved := false;

  if Axioms then
  begin
    i := 0;
    OutputLines.Add('______________________________');
    OutputLines.Add('Source formulas (axioms) : ');
    while i < ListF.Count do
    begin
      s := '';
      s := s+'F'+IntToStr(i);

      if ListF.Items[i] <> nil then
      begin
        if TObject(ListF.Items[i]) is TQuery then s := s+' ('+cneg+'query)';
        s := s+' : ';
        TFALFormula(ListF.Items[i]).Print(s);
      end;
      OutputLines.Add(s);
      inc(i);
      Application.ProcessMessages;
      if retf then begin Break; end;
    end;
    OutputLines.Add('______________________________');
  end
  else
  begin
    s := 'Solution to ';
    TFALFormula(ListF.Items[ListF.Count-1]).Print(s);
    OutputLines.Add(s);
  end;
  { Linear strategy going to the all axioms first. }
  if Strategy = SrtModifLinear then
  begin
    i := ListF.Count-2;
    while i >= 0 do
    begin
       rsv1 := 'F'+IntToStr(i);
       if (ListF.Items[i] <> nil) then
         TFALFormula(ListF.Items[i]).AllPosRes4;
       dec(i);
       Application.ProcessMessages;
       if retf then begin Break; end;
    end;
    rsv1 := 'F'+IntToStr(ListF.Count-1);
    if retf = false then TFALFormula(ListF.Items[ListF.Count-1]).AllPosRes4;
  end
  else
  { Linear strategy. }
  if Strategy = SrtLinear then
  begin
    rsv1 := 'F'+IntToStr(ListF.Count-1);
    TFALFormula(ListF.Items[ListF.Count-1]).AllPosRes4;
  end
  else
  { Breadth-first strategy. }
  if Strategy = SrtBreadth then
  begin
    i := ListF.Count-1;
    while i >= 0 do
    begin
       rsv1 := 'F'+IntToStr(i);
       if (ListF.Items[i] <> nil) then
         TFALFormula(ListF.Items[i]).AllPosRes3;
       dec(i);
       Application.ProcessMessages;
       if retf then begin Break; end;
    end;
  while (infer)and(retf = false) do
  begin
    infer := false;
    lastr1 := rsvnum;
    i := ListR.Count-1;
    while i >= lastr2 do
    begin
      rsv1 := 'R'+IntToStr(i);
      if (ListR.Items[i] <> nil) then
      TFALFormula(ListR.Items[i]).AllPosRes;
      dec(i);
      Application.ProcessMessages;
      if retf then begin Break; end;
    end;

    lastr2 := lastr1;
    i := ListF.Count-1;
    while i >= 0 do
    begin
       rsv1 := 'F'+IntToStr(i);
       if (ListF.Items[i] <> nil) then
         TFALFormula(ListF.Items[i]).AllPosRes2;
       dec(i);
       Application.ProcessMessages;
       if retf then begin Break; end;
    end;
  end;
  end;
  if Resolvents then
  begin
    i := 0;
    OutputLines.Add('_____________________');
    OutputLines.Add('Produced resolvents : ');
    while i < ListR.Count do
    begin
      s := '';
      s := s+'R'+IntToStr(i)+' : ';
      if ListR.Items[i] <> nil then
      begin
        TFALFormula(ListR.Items[i]).Print(s);
        if TFALFormula(ListR.Items[i]).DisabledDCF then
          s := s + '(DISABLED BY DCF)';
      end;
      OutputLines.Add(s);
      inc(i);
      Application.ProcessMessages;
      if retf then begin Break; end;
    end;
    OutputLines.Add('_____________________');
  end;
  if Stat then
  begin
    OutputLines.Add('Inference statistics : ');
    OutputLines.Add('Resolvents : '+IntToStr(ListR.count));
    OutputLines.Add('Removed tautologies : '+IntToStr(ignnum));
    OutputLines.Add('DCF consequent resolvents : '+IntToStr(nummol));
    OutputLines.Add('DCF killed resolvents : '+IntToStr(numkill));
    OutputLines.Add('Unification : '+IntToStr(numsim));
    OutputLines.Add('Simplification : '+IntToStr(numsimX));
    OutputLines.Add('_____________________');
  end;
  if UsedMem > MaxMem then MaxMem:=UsedMem;
  ClearResolvents;
end;

function FSimplifyH(o : TObject; var tp : char; argums : array of const)
                                                       : TObject; forward;

var truthstr : string;

{  AllPosRes procedures try to resolve with all possible formulas by
   different ways. }
procedure TFALFormula.AllPosRes;
var i, ad : longint;
begin
    i := Owner.ListF.Count-1;
    while i >= 0 do
    begin
      rsv2 := 'F'+IntToStr(i);
      if Owner.ListF.Items[i] <> nil then
        self.TryToRes2(TFALFormula(Owner.ListF.Items[i]));
      if retf then begin Exit; end;
      dec(i);
    end;
    i := Owner.ListR.Count-1;
    while i >= 0 do
    begin
      rsv2 := 'R'+IntToStr(i);
      if Owner.ListR.Items[i] <> nil then
      self.TryToRes2(TFALFormula(Owner.ListR.Items[i]));
      if retf then begin Exit; end;
      dec(i);
    end;
end;

procedure TFALFormula.AllPosRes2;
var i, ad : longint;
begin
    i := Owner.ListR.Count-1;
    while i >= 0 do
    begin
      rsv2 := 'R'+IntToStr(i);
      if Owner.ListR.Items[i] <> nil then
      self.TryToRes2(TFALFormula(Owner.ListR.Items[i]));
      if retf then begin Exit; end;
      dec(i);
    end;
end;

procedure TFALFormula.AllPosRes3;
var i, ad : longint;
begin
    und := true;
    i := Owner.ListF.Count-1;
    while i >= 0 do
    begin
      rsv2 := 'F'+IntToStr(i);
      if Owner.ListF.Items[i] <> nil then
        self.TryToRes2(TFALFormula(Owner.ListF.Items[i]));
      if retf then begin Exit; end;
      dec(i);
    end;
    und := false;
end;

procedure TFALFormula.AllPosRes4;
var i, ad : longint;
begin
    i := Owner.ListF.Count-1;
    while i >= 0 do
    begin
      rsv2 := 'F'+IntToStr(i);
      if Owner.ListF.Items[i] <> nil then
        self.TryToRes3(TFALFormula(Owner.ListF.Items[i]));
      if retf then begin Exit; end;
      dec(i);
    end;
    i := Owner.ListR.Count-1;
    while i >= 0 do
    begin
      rsv2 := 'R'+IntToStr(i);
      if Owner.ListR.Items[i] <> nil then
      self.TryToRes3(TFALFormula(Owner.ListR.Items[i]));
      if retf then begin Exit; end;
      dec(i);
    end;
end;


var Ancs : TObject;

{ This function checks, if one formula is a child of the second one
  or vice-versa. }
function TFALFormula.IsValidPair;
var ElPar : TFALFormula;
  function ValPar(X : TFALFormula) : boolean;
  begin
    Result := false;
    if X = ElPar then Result := true;
    if (Result = false) then
    begin
      if X.Parent1 <> nil then Result := ValPar(X.Parent1);
      if (Result = false)and(X.Parent2 <> nil) then
         Result := ValPar(X.Parent2);
    end;
  end;
begin
  Result := false;
  if ((Parent1 = nil)or(X.Parent1 = nil)) then Result := true
  else
  begin
    ElPar := self;
    Result := ValPar(X);
    if Result = false then
    begin
      ElPar := X;
      Result := ValPar(self);
    end;
  end;
end;


{ Tries to resolve the formula with the second one, simplifies it and
 do other necessary transformations. }
function TFALFormula.TryToRes;
var T1, T2, TT1, TT2, TP1, TP2 : TAtom; FN1, FN2 : TFALFormula;
    ad : integer; ind : integer; premise : boolean;
begin
if (Owner.Checkby2) then
  if self.DisabledDCF or X.DisabledDCF then
  begin
    Result := 0;
    Exit;
  end;

if (Owner.Without1)
  or((Owner.Filtra1)and(IsValidPair(X)))
  or((Owner.Support1)and(Support or X.Support))
then
begin
   mgu1.clear;
   FN1 := self; FN2 := X;
   TT1 := FN1.MF; TT2 := FN2.MF; result := 0; ad := -1;
   if (TT1 <> nil)and(TT2 <> nil) then
   begin
    while (TT1 <> nil) do
    begin
      if result = 0 then TT2 := FN2.MF;
      result := 0;
      while (TT2 <> nil) do
      begin
        Application.ProcessMessages;
        if Owner.stopf then Exit;
        result := TT1.UnifyS(TT2);    { While MGU is not founded. }
        if result <> 0
          then
        begin
          T1 := TT1; T2 := TT2;
          if T1.Pol = 0 then
          begin
            if T2.Pol = -1 then premise := true else premise := false;
          end
          else if T1.Pol = 1 then premise := true else premise := false;
          FN1 := FN1.Resolve(FN2, T1, T2, premise);
          if Owner.Unsim then
          begin
            truthstr := '['+rsv1+'&'+rsv2+']'+' : ';
            if FN1.Cont <> nil then FN1.Print(truthstr);
            Owner.OutputLines.Add(truthstr);
          end;
          if (FN1 <> self) then
          begin
            FN1.Simplify;
            FN1.Cont.EvalPol;
            FN1 := FN1.InsertToR;   { Checks redundancy. }
            if (FN1 <> nil) then
            begin
              if (FN1.Cont.Ev <> 0) then
                begin
                  if (FN1.Cont.Ev = -1)
                  then
                  begin
                    PrintRefutation(FN1);
                  end
                  else inc(ignnum);
                  FN1.Destroy;
                end
                else
                begin
                    if Owner.Progress then
                    begin
                      truthstr := '';
                      if Owner.Sources then
                         truthstr := 'R'+IntToStr(rsvnum)
                                  +' ['+rsv1+'&'+rsv2+']'+' : ';
                      FN1.Print(Truthstr);
                      Owner.OutputLines.Add(truthstr);
                    end;
                    Owner.ListR.Add(FN1);
                    //DCF Killing method
                    FN1.DCFKillEqv;

                    inc(rsvnum); infer := true;
                end;
            end
          end;
          if Owner.ExitOn1 then Exit;
          { if requested, it continues to find next MGU. }
        end;
        if Owner.ExitOn3 then Exit;
        FN1 := self; FN2 := X;
        TT2 := TAtom(TT2.R); result := 0;
      end;
      if result = 0 then TT1 := TAtom(TT1.R);
    end;
 end;
end;
end;

function TFALFormula.TryToRes2;
var T1, T2, TT1, TT2, TP1, TP2 : TAtom; FN1, FN2 : TFALFormula;
    ad : integer; ind : integer; premise : boolean;
begin

if (Owner.Checkby2) then
  if self.DisabledDCF or X.DisabledDCF then
  begin
    Result := 0;
    Exit;
  end;

if (Owner.Without1)
  or((Owner.Filtra1)and(IsValidPair(X)))
  or((Owner.Support1)and(Support or X.Support))
then
begin
   mgu1.clear;
   FN1 := self; FN2 := X;
   TT1 := FN1.MF; TT2 := FN2.MF; result := 0; ad := -1;
   if (TT1 <> nil)and(TT2 <> nil) then
   begin
    while (TT1 <> nil) do
    begin
      if result = 0 then TT2 := FN2.MF;
      result := 0;
      while (TT2 <> nil) do
      begin
        Application.ProcessMessages;
        if Owner.stopf then Exit;
        result := TT1.UnifyS(TT2);
        if result <> 0
          then
        begin
          T1 := TT1; T2 := TT2;
          if T1.Pol = 0 then
          begin
            if T2.Pol = -1 then premise := true else premise := false;
          end
          else if T1.Pol = 1 then premise := true else premise := false;
          FN1 := FN1.Resolve(FN2, T1, T2, premise);
          if (FN1 <> self) then
          begin
             if Owner.Unsim then
             begin
               truthstr := '['+rsv1+'&'+rsv2+']'+' : ';
               if FN1.Cont <> nil then FN1.Print(truthstr);
               Owner.OutputLines.Add(truthstr);
             end;
             FN1.Simplify;
             FN1.Cont.EvalPol;
             FN1 := FN1.InsertToR;   { Checks redundancy. }
             if (FN1 <> nil) then
             begin
               if (FN1.Cont.Ev <> 0) then
               begin
                 if (FN1.Cont.Ev = -1)
                 then
                 begin
                   PrintRefutation(FN1);
                 end
                 else inc(ignnum);
                 FN1.Destroy;
               end
               else
               begin
                 if Owner.Progress then
                 begin
                   truthstr := '';
                   if Owner.Sources then
                      truthstr := 'R'+IntToStr(rsvnum)
                               +' ['+rsv1+'&'+rsv2+']'+' : ';
                   FN1.Print(Truthstr);
                   Owner.OutputLines.Add(truthstr);
                 end;
                 Owner.ListR.Add(FN1);

                 //DCF Killing method
                    FN1.DCFKillEqv;

                 inc(rsvnum); infer := true;
               end;
             end;
             if Owner.ExitOn1 then Exit;
          end;
        end;
        if Owner.ExitOn3 then Exit;
        FN1 := self; FN2 := X;
        TT2 := TAtom(TT2.R); result := 0;
      end;
      if result = 0 then TT1 := TAtom(TT1.R);
    end;
 end;
end;
end;

function TFALFormula.TryToRes3;
var T1, T2, TT1, TT2, TP1, TP2 : TAtom; FN1, FN2 : TFALFormula; ad : integer; ind : integer;
    rsvbf : string; premise : boolean;
begin

if (Owner.Checkby2) then
  if self.DisabledDCF or X.DisabledDCF then
  begin
    Result := 0;
    Exit;
  end;

if (Owner.Without1)
  or((Owner.Filtra1)and(IsValidPair(X)))
  or((Owner.Support1)and(Support or X.Support))
then
begin
   mgu1.clear;
   rsvbf := rsv1;
   FN1 := self; FN2 := X; 
   TT1 := FN1.MF; TT2 := FN2.MF; result := 0; ad := -1;
   if (TT1 <> nil)and(TT2 <> nil) then
   begin
    while (TT1 <> nil) do
    begin
      if result = 0 then TT2 := FN2.MF;
      result := 0;
      while (TT2 <> nil) do
      begin
        Application.ProcessMessages;
        if Owner.stopf then Exit;
        result := TT1.UnifyS(TT2);
        if result <> 0
          then
        begin
          T1 := TT1; T2 := TT2;
          if T1.Pol = 0 then
          begin
            if T2.Pol = -1 then premise := true else premise := false;
          end
          else if T1.Pol = 1 then premise := true else premise := false;
          FN1 := FN1.Resolve(FN2, T1, T2, premise);
          if (FN1 <> self) then
          begin
             if Owner.Unsim then
             begin
               truthstr := '['+rsv1+'&'+rsv2+']'+' : ';
               if FN1.Cont <> nil then FN1.Print(truthstr);
               Owner.OutputLines.Add(truthstr);
             end;
             FN1.Simplify;
             FN1.Cont.EvalPol;
             FN1 := FN1.InsertToR;   { Checks redundancy. }
             if (FN1 <> nil) then
             begin
               if (FN1.Cont.Ev <> 0) then
               begin
                 if (FN1.Cont.Ev = -1)
                 then
                 begin
                   PrintRefutation(FN1);
                 end
                 else inc(ignnum);
                 FN1.Destroy;
               end
               else
               begin
                 if Owner.Progress then
                 begin
                   truthstr := '';
                   if Owner.Sources then
                      truthstr := 'R'+IntToStr(rsvnum)
                               +' ['+rsv1+'&'+rsv2+']'+' : ';
                   FN1.Print(Truthstr);
                   Owner.OutputLines.Add(truthstr);
                 end;
                 Owner.ListR.Add(FN1);

                 //DCF Killing method
                    FN1.DCFKillEqv;

                 inc(rsvnum); infer := true;
                 rsv1 := 'R'+IntToStr(rsvnum-1);
                 FN1.AllPosRes4;
                 rsv1 := rsvbf;
               end;
             end;
            if Owner.ExitOn1 then Exit;
          end;
        end;
        if Owner.ExitOn3 then Exit;
        FN1 := self; FN2 := X;
        TT2 := TAtom(TT2.R); result := 0;
      end;
      if result = 0 then TT1 := TAtom(TT1.R);
    end;
 end;
end;
end;

procedure TFALFormula.MakeSymbolic;
begin
  if Symbolic='' then TSub(Cont).RPrint(Symbolic);
end;

function TFALFormula.TrivialCheck(X : TFALFormula):boolean;
begin
  if (Symbolic <>'')and(X.Symbolic<>'') then Result := Symbolic = X.Symbolic
  else Result := False; 
end;

procedure TFALFormula.DCFKill;
var TX, TC1, TC2, TC3 : TFALFormula; i : longint;
begin
    i := 0;
    TX := self;
    if (Owner.Checkby2) then
    while i < Owner.ListR.Count do
    begin
      if (Owner.ListR.Items[i] <> nil)and(Owner.ListR.Items[i] <> TX) then
      begin
        TC2 := TFALFormula(Owner.ListR.Items[i]);

          if TC2.DisabledDCF then
          begin
            inc(i);
            continue;
          end;

        TC1 := TX.CreateVF(TC2, TImp);
        TC1.Cont.neg := true;
        TC1.Cont.EvalPol;


        TC3 := TC1.OptimizeX;


  { VF is totally optimized and it's checked for inconsistency. }
        TC1.DestroyVF(TX, TC2);
        if (TC3.Cont.Ev = -1) then
        begin
              truthstr := '';
              if Owner.SRcheck1 then
              begin
                truthstr := 'DCF Kill: ';
                TX.Print(truthstr);
                truthstr := truthstr + 'R' + IntToStr(i)+' is consequent;';
                if TC2.Degree > TX.Degree then
                  truthstr := truthstr + 'but degree of R'+IntToStr(i)+' > ' + FloatToStr(TX.Degree)
                  else truthstr := truthstr + '(DISABLED)';
                Owner.OutputLines.Add(truthstr);
              end;
              if TC2.Degree > TX.Degree then
              begin
                inc(i);
                TC3.Free;
                continue;
              end;
              TC2.DisabledDCF := true;
              inc(numKill);
        end;
        TC3.Free;
      end;
      inc(i);
      Application.ProcessMessages;
      if Owner.stopf then Exit;
    end;
end;

procedure TFALFormula.DCFKillEqv;
var TX, TC1, TC2, TC3 : TFALFormula; i : longint;
begin
    i := 0;
    TX := self;
    if (Owner.Checkby2) then
    while i < Owner.ListR.Count do
    begin
      if (Owner.ListR.Items[i] <> nil)and(Owner.ListR.Items[i] <> TX) then
      begin
          TC2 := TFALFormula(Owner.ListR.Items[i]);

          if TC2.DisabledDCF then
          begin
            inc(i);
            continue;
          end;

        TC1 := TX.CreateVF(TC2, TImp);
        TC1.Cont.neg := true;
        TC1.Cont.EvalPol;


        TC3 := TC1.OptimizeX;


  { VF is totally optimized and it's checked for inconsistency. }
        TC1.DestroyVF(TX, TC2);
        if (TC3.Cont.Ev = -1) then
        //must check also opposite implication
          begin
            if TC2.Degree < TX.Degree then
              begin
                truthstr := '';
                if Owner.SRcheck1 then
                begin
                  truthstr := 'DCF Kill: ';
                  TX.Print(truthstr);
                  truthstr := truthstr + '//consequent to R' + IntToStr(i)+'(DISABLED);';
                  Owner.OutputLines.Add(truthstr);
                end;
                TC3.Free;
                inc(i);
                inc(numKill);
                TC2.DisabledDCF := true; continue;
              end;

            TC3.Free;
            TC1 := TC2.CreateVF(TX, TImp);
            TC1.Cont.neg := true;
            TC1.Cont.EvalPol;
            TC3 := TC1.OptimizeX;
            TC1.DestroyVF(TC2, TX);
            if (TC3.Cont.Ev = -1) then
            begin
              truthstr := '';
              if Owner.SRcheck1 then
              begin
                truthstr := 'DCF Kill: ';
                TX.Print(truthstr);
                truthstr := truthstr + 'R' + IntToStr(i)+' is equivalent;';
                if TC2.Degree > TX.Degree then
                  truthstr := truthstr + 'but degree of R'+IntToStr(i)+' > ' + FloatToStr(TX.Degree)
                  else truthstr := truthstr + '(DISABLED)';
                Owner.OutputLines.Add(truthstr);
              end;
              if TC2.Degree > TX.Degree then
              begin
                inc(i);
                TC3.Free;
                continue;
              end;
              TC2.DisabledDCF := true;
              inc(numKill);
            end;
          end;

        TC3.Free;
      end;
      inc(i);
      Application.ProcessMessages;
      if Owner.stopf then Exit;
    end;
end;


  { Evaluates infix operators. }
function EvalBuilt(o : TObject; var tp : char; argums : array of const)
                                                              : TObject;
var r : extended; N0, N1 : TObject; fc : char;
begin
  Result := o;
  if o is TFunctor then
  begin
    if (TFunctor(o).Functor[1] in infixoper) then
    begin
     N0 := (TFunctor(o).Params.Items[0]);
     N1 := (TFunctor(o).Params.Items[1]);
     while N0 is TVariable do
        N0 := TVariable(N0).Id;
     while N1 is TVariable do
        N1 := TVariable(N1).Id;
     if (N0 is TNumber)and(N1 is TNumber) then
     begin
       try
       case TFunctor(o).Functor[1] of
       '+' : r := TNumber(N0).GetNumber +
              TNumber(N1).GetNumber;
       '-' : r := TNumber(N0).GetNumber -
              TNumber(N1).GetNumber;
       '*' : r := TNumber(N0).GetNumber *
              TNumber(N1).GetNumber;
       '/' : r := TNumber(N0).GetNumber /
              TNumber(N1).GetNumber;
       end;
       except
         on EMathError do Exit;
       end;
       TFunctor(o).RFree;
       Result := TReal.CreateN(r);
     end;
    end;
  end;
end;



  { Simplification procedure. }
function FSimplifyH(o : TObject; var tp : char; argums : array of const)
                                                               : TObject;
var i : longint; x1, x2, xa : shortint; negf : shortint;
begin
  if (TSub(o).L <> nil)and(TSub(o).R <> nil)
    and((TSub(o).L.Ev <> 0)or(TSub(o).R.Ev <> 0)) then
  begin
    inc(numSimX);
    x1 := TSub(o).L.Ev; x2 := TSub(o).R.Ev;
    if o is TCon then xa := Min(x1, x2)
    else if o is TDis then xa := Max(x1, x2)
    else if o is TImp then xa := Max(-x1, x2);
    if (xa <> 0) then
      begin
        TSub(o).L.RFree; TSub(o).L := nil;
        TSub(o).R.RFree; TSub(o).R := nil;
        TSub(o).Ev := xa;
        { The case, that leads to complete evaluation of a node
          e.g. (x and true) leads to x .}
      end
    else if (xa = 0) then
      begin
        negf := 0;
        if ((o is TImp)and(x2 = -1)) then negf := 1;
        if x1 = 0 then o := TSub(o).L.SubstTo else o := TSub(o).R.SubstTo;
        if negf <> 0 then
        begin
          TSub(o).negate;
        end;
        { The case, that leads to a partial evaluation of a node.
          e.g. (x and false) leads to false .}
      end;
    if TSub(o).neg then TSub(o).Ev := -TSub(o).Ev;
  end;
  Result := o;
end;

procedure UndoMgu;
begin
  mgu1.clear;
end;

{ Preparation when the MGU was found. }
procedure PrepareMgu;
var i : longint; var X : TVariable;
begin
  i := 0;
  while i < mgu1.Count do
  begin
    X := TVariable(mgu1.Items[i]);
    X.NewId := TId(mgu1.Items[i+1]); { Assigns substitution. }
    {fix}
    if X.NewId is TVariable then TVariable(X.NewId).IsInMgu := true;
    i := i+2;
  end;
  mgu1.clear;
end;

var invert : boolean;

function TFALFormula.Resolve;
var F1, F2, F3 : TSub; s : string; FL : TFALFormula; ad : integer;
    Pol1, polar1 : boolean; var TT1, TT2 : TAtom; rs : integer;tr1 : shortint;
begin
 Result := self;
 begin
  FirstMol := nil; LastMol := nil; 

  NewQ := nil;
  NewQ := TQuant.Create;
  F3 := TDis.Create;        { New disjunction is prepared. }
  FL := TFALFormula.Create(Owner);
  FL.Degree := LukMult(self.Degree,X.Degree);

    if (self = nil)and(X = nil) then
    begin
      FL.Parent1 := nil;
      FL.Parent2 := nil;
    end
    else
    begin
      FL.Parent1 := self;
      FL.Parent2 := X;
    end;

  if Support or X.Support then FL.Support := true;
  if premise then tr1 := -1 else tr1 := 1;


  if self <> X then
  begin
    if T1.neg then T1.Ev := -tr1 else T1.Ev := tr1;
    if self <> X then if T2.neg then T2.Ev := tr1 else T2.Ev := -tr1;
    TT1 := T1;
    polar1 := false;
    if (TT1 <> nil) then
    begin
        TT2 := MF; rs := 0;
        while (TT2 <> nil) do
        begin
          rs := TT1.UnifyS(TT2);
          if rs = 0 then TT2 := TAtom(TT2.R)
          else
          begin
            if polar1 then TT2.Ev := -tr1 else TT2.Ev := tr1;
            if TT2.neg then TT2.Ev := -TT2.Ev;
            rs := 0;
            TT2 := TAtom(TT2.R);
          end;
        end;
    end;
    if (self <> X)and
      (Owner.General) then
    begin
      if T2.neg then T2.Ev := tr1 else T2.Ev := -tr1;
      polar1 := true;
      if (TT1 <> nil) then
      begin
          TT2 := X.MF; rs := 0;
          while (TT2 <> nil) do
          begin
            rs := TT1.UnifyS(TT2);
            if rs = 0 then TT2 := TAtom(TT2.R)
            else
            begin
              if polar1 then TT2.Ev := -tr1 else TT2.Ev := tr1;
              if TT2.neg then TT2.Ev := -TT2.Ev;
              TT2 := TAtom(TT2.R);
              rs := 0;
            end;
          end;
      end;
    end;
  end
  else CompleteUnify(T1);
  PrepareMgu;
  FL.Cont := F3; F3.Ac := FL;
  Ancs := F3; 
  F1 := Cont.Dive('r', nil, [nil]);   { Makes a copy of the first formula
                                        with substituted variables. }
  Ancs := F3;  
  if (self = X)
   then
  begin
    invert := true;
    F2 := X.Cont.Dive('r', nil, [nil]);
    invert := false;
  end
  else F2 := X.Cont.Dive('r', nil, [nil]);
  F3.L := F1; F3.R := F2; FL.MF := FirstMol; FL.ML := LastMol;
  if NewQ.Count > 0 then  FL.Cont.AppendToQuant(NewQ)
  else NewQ.Destroy;  { Assigned variables have to be carefully treated. }
                      { They are quantified in the highest node. }
  Dive('d', nil, [nil]);
  if self <> X then
  begin
    X.Dive('d', nil, [nil]);  { Performs erasure of temporal items. }
  end;
  if FL.MF <> nil  then FL.MF.L := nil;
  if FL.ML <> nil  then FL.ML.R := nil;
  result := FL;
 end;
end;




procedure TFALFormula.CompleteUnify;
var rsunif  : integer; lastNew, At1, At2 :TAtom; unified : boolean;
    tr1 : shortint; premise : boolean;
begin

  At1:=T1;
  if premise then tr1 := -1 else tr1 := 1;
  if At1.neg then At1.TempEv := -tr1 else At1.TempEv := tr1;

  repeat

    unified := false;
    while At1 <> nil do
    begin
      if (At1.TempEv = 0) then
      begin At1 := TAtom(At1.R); continue; end;
      At2 := MF;

      while At2 <> nil do
      begin
        if (At1.TempEv <> 0)and(At2.TempEv <> 0) then
          begin At2 := TAtom(At2.R); continue; end;
        rsunif := 0;
        rsunif := At1.UnifyS(At2);
        if rsunif <> 0 then
        begin
          At2.TempEv := tr1;
          if At2.neg then At2.TempEv := -At2.TempEv;
          unified := true;
        end;
        At2 := TAtom(At2.R);
      end;


    At1 := TAtom(At1.R);
    Undomgu;
    end;
    At1 := MF;
  until not unified;

  At2 := MF;

      while At2 <> nil do
      begin
        At2.Ev := At2.TempEv;
        At2 := TAtom(At2.R);
      end;


end;


function TFALFormula.ResolveAWCFirst;
var F1, F2, F3 : TSub; s : string; FL : TFALFormula; ad : integer;
    Pol1 : boolean; var TT1, TT2 : TAtom; rs : integer;
begin
  Result := self;
  FirstMol := nil; LastMol := nil; 
  
  NewQ := nil;
  NewQ := TQuant.Create;
  F3 := TDis.Create;        { New disjunction is prepared. }
  FL := TFALFormula.Create(Owner);
  FL.Degree := LukMult(self.Degree,X.Degree);

  CompleteUnify(T1);

  PrepareMgu;
  FL.Cont := F3; F3.Ac := FL;
  Ancs := F3;
  F1 := Cont.Dive('r', nil, [nil]);   { Makes a copy of the first formula
                                        with substituted variables. }
  Ancs := F3;

  invert := true;
  F2 := X.Cont.Dive('r', nil, [nil]);
  invert := false;

  F3.L := F1; F3.R := F2; FL.MF := FirstMol; FL.ML := LastMol;
  if NewQ.Count > 0 then  FL.Cont.AppendToQuant(NewQ)
  else NewQ.Destroy;  { Assigned variables have to be carefully treated. }
                      { They are quantified in the highest node. }
  Dive('d', nil, [nil]);

  if FL.MF <> nil  then FL.MF.L := nil;
  if FL.ML <> nil  then FL.ML.R := nil;
  result := FL;

end;


function TFALFormula.ResolveAWC;
var F1, F2, F3 : TSub; s : string; FL : TFALFormula; ad : integer;
    Pol1 : boolean; var TT1, TT2 : TAtom; rs : integer;
begin
  Result := self;
  FirstMol := nil; LastMol := nil; 

  NewQ := nil;
  NewQ := TQuant.Create;
  F3 := TDis.Create; FL := self;
  FL.Degree := LukMult(self.Degree,X.Degree);

  CompleteUnify(T1);

  F1 := FL.Cont; FL.Cont.Ac := F3;
  FirstMol := FL.MF; LastMol := FL.ML;
  PrepareMgu;
  Ancs := F3;

  invert := true;
  F2 := X.Cont.Dive('r', nil, [nil]);
  invert := false;

  FL.Cont := F3; F3.Ac := FL;
  F3.L := F1; F3.R := F2; FL.MF := FirstMol; FL.ML := LastMol;
  if NewQ.Count > 0 then FL.Cont.AppendToQuant(NewQ) else NewQ.Destroy;

  if FL.MF <> nil  then FL.MF.L := nil;
  if FL.ML <> nil  then FL.ML.R := nil;
  result := FL;

end;




var dest : boolean;

{ Function marks variable's used property,
                 if it is used in some expression. }
function UsedVars(o : TObject; var tp : char; argums : array of const)
                                                             : TObject;
var i : longint; X : TQuant;
begin
  if o is TSub then
  begin
    if TSub(o).Q <> nil then
      begin
        X := TSub(o).Q;
        for i := 0 to X.Count-1 do
        begin
          if (X.Items[i] <> nil)and(TVariable(X.Items[i]).Id <> nil) then
          begin
            TVariable(X.Items[i]).Id.Dive('3', UsedVars, [nil]);
          end;
        end;
      end;
  end;
  if o is TAtom then tp := '3'
  else if o is TVariable then
  begin
    TVariable(o).Used := true;
  end;
  Result := o;
end;

{ Simplification of the formula. }
procedure TFALFormula.Simplify;
begin
  Dive('3', EvalBuilt, [nil]);   { Evaluates infix operators. }
  Firstmol := MF; LastMol := ML;
  Dive('s', FSimplifyH, [nil]);  { Uses logical rules to log. const. }
  MF := FirstMol; ML := LastMol;
  begin
   dest := true;
   while dest do
   begin
     dest := false;
     Dive('4', UsedVars, [nil]);  { Removes of unused variables. }
     Dive('d', nil, [nil]);
   end;
  end;
end;


function TFALFormula.OptimizeX;
var F1, F2, F3 : TSub; s : string; FL : TFALFormula; ad : integer;
    Pol1 : boolean; var T1, TT1, TT2 : TAtom; rst : integer;
    FM1, Trash : TFALFormula; premise : boolean;
    anc : TSub;
begin
  FM1 := self; premise := false;rst := 0;
  slu:=true;
  TT1 := FM1.MF;
  while (TT1 <> nil)and(rst<Owner.DCF) do  //resolves until formula is empty.
  begin
     TT2 := TT1;
     T1 := TT1;

     if T1.Pol = 1 then premise := true else premise := false;

     Trash := FM1;
           //truthstr := ':';
           //FM1.Print(truthstr);
           //Owner.OutputLines.Add(truthstr);
     if FM1 <> self then
       FM1 := FM1.ResolveAWC(FM1, T1, premise)
     else FM1 := FM1.ResolveAWCFirst(FM1, T1, premise);
           //truthstr := 'After: ';
           //FM1.Print(truthstr);
           //Owner.OutputLines.Add(truthstr);
     FM1.Simplify;
     FM1.Cont.EvalPol;
     inc(rst);
     TT1 := FM1.MF;
  end;
  slu:=false;
  Result := FM1;
end;


{ Creates virtual formula - link to two physical formulas by a connective. }
function TFALFormula.CreateVF;
begin
  Result := TFALFormula.Create(Owner);
  Result.Cont := T.Create;
  Result.Cont.Ac := Result;
  Result.Cont.R := X.Cont;
  Result.Cont.L := Cont;
  Result.Cont.L.Ac := Result.Cont;
  Result.Cont.R.Ac := Result.Cont;
  Cont := nil; X.Cont := nil;
  if MF <> nil then Result.MF := MF else Result.MF := X.MF;
  if X.ML <> nil then Result.ML := X.ML else Result.ML := ML;
  if ML <> nil then ML.R := X.MF;
  if X.MF <> nil then X.MF.L := ML;
  Result.Cont.EvalPol;
end;

procedure TFALFormula.DestroyVF;
begin
  X1.Cont := Cont.L;
  X2.Cont := Cont.R;
  X1.Cont.Ac := X1;
  if X1.ML <> nil then X1.ML.R := nil;
  X2.Cont.Ac := X2;
  if X2.MF <> nil then X2.MF.L := nil;
  Cont.L := nil; Cont.R := nil;
  X1.Cont.EvalPol;
  X2.Cont.EvalPol;
  Free;
end;

{ Checks, if formula is redundant with respect to other resolvents. }
function TFALFormula.InsertToREqv;
var TX, TC1, TC2, TC3 : TFALFormula; i : longint; TT1 : TAtom;

begin
  Result := self;
  TX := OptimizeX;
  if TX.Cont.Ev = -1 then
  begin
  if TX <> self then
    begin
      Free;
    end;
    Result := TX;
    Exit;
  end;


  i := 0;
  if TX <> self then
    begin
      TX.Free;
    end;
  TX := self;

    if Cont<>nil then Cont.EvalPol;

    if (TX.Cont.Ev <> 0) then Exit;

    MakeSymbolic;
    while i < Owner.ListR.Count do
      begin
        if (Owner.ListR.Items[i] <> nil) then
        begin
          if (TFALFormula(Owner.ListR.Items[i])).DisabledDCF then
          begin
            inc(i);continue;
          end;

          TC2 := TFALFormula(Owner.ListR.Items[i]);

          if TX.TrivialCheck(TC2) then
          begin
            truthstr := '';
            if Owner.SRcheck1 then
            begin
                  truthstr := 'DCF: ';
                  TX.Print(truthstr);
                  truthstr := truthstr + '//identical to R' + IntToStr(i)+';';
            end;
            if TC2.Degree < TX.Degree then
              begin
                TC2.Degree := TX.Degree;
                truthstr := truthstr + 'degree of R'+IntToStr(i)+' raised to: ' + FloatToStr(TX.Degree);
              end;
                if truthstr<>'' then Owner.OutputLines.Add(truthstr);

              Free; inc(nummol);
              Result := nil; Exit;
          end;
        end;
        inc(i);
      end;

     i := 0;

    if (TX.Cont.Ev = 0)and(Owner.RedCheck1 = false) then
    begin
      while i < Owner.ListR.Count do
      begin
        if (Owner.ListR.Items[i] <> nil) then
        begin
          if (TFALFormula(Owner.ListR.Items[i])).DisabledDCF then
          begin
            inc(i);continue;
          end;

          TC2 := TFALFormula(Owner.ListR.Items[i]);
          TC1 := TC2.CreateVF(TX, TImp);
          TC1.Cont.neg := true;
          TC1.Cont.EvalPol;

          TC3 := TC1.OptimizeX;

          TC1.DestroyVF(TC2, TX);

          if (TC3.Cont.Ev = -1) then
          //must check also opposite implication
          begin
            if TC2.Degree >= TX.Degree then
              begin
                truthstr := '';
                if Owner.SRcheck1 then
                begin
                  truthstr := 'DCF: ';
                  TX.Print(truthstr);
                  truthstr := truthstr + '//consequent to R' + IntToStr(i)+';';
                  Owner.OutputLines.Add(truthstr);
                end;
                TC3.Free;
                Free; inc(nummol);
                Result := nil; Exit;
              end;

            TC3.Free;
            TC1 := TX.CreateVF(TC2, TImp);
            TC1.Cont.neg := true;
            TC1.Cont.EvalPol;
            TC3 := TC1.OptimizeX;
            TC1.DestroyVF(TX, TC2);
            if (TC3.Cont.Ev = -1) then
            begin
              truthstr := '';
              if Owner.SRcheck1 then
              begin
                truthstr := 'DCF: ';
                TX.Print(truthstr);
                truthstr := truthstr + '//equivalent to R' + IntToStr(i)+';';
              end;
              if TC2.Degree < TX.Degree then
              begin
                TC2.Degree := TX.Degree;
                truthstr := truthstr + 'degree of R'+IntToStr(i)+' raised to: ' + FloatToStr(TX.Degree);
              end;
                if truthstr<>'' then Owner.OutputLines.Add(truthstr);
              TC3.Free;
              Free; inc(nummol);
              Result := nil; Exit;
            end;
          end
          else
          begin
            TC3.Free;
            inc(i);
            continue;
          end;
        end;
        TC3.Destroy;
        inc(i);
        Application.ProcessMessages;
        if Owner.stopf then Exit;
      end;

    end;

    Result := TX;

  if UsedMem > MaxMem then MaxMem:=UsedMem;
end;

function TFALFormula.InsertToR;
var TX, TC1, TC2, TC3 : TFALFormula; i : longint; TT1 : TAtom;

begin
  Result := self;
  TX := OptimizeX;
  if TX.Cont.Ev = -1 then
  begin
  if TX <> self then
    begin
      Free;
    end;
    Result := TX;
    Exit;
  end;


  i := 0;
  if TX <> self then
    begin
      TX.Free;
    end;
    TX := self;

    if Cont<>nil then Cont.EvalPol;

    if (TX.Cont.Ev <> 0) then Exit;

    MakeSymbolic;
    while i < Owner.ListR.Count do
      begin
        if (Owner.ListR.Items[i] <> nil) then
        begin
          if (TFALFormula(Owner.ListR.Items[i])).DisabledDCF then
          begin
            inc(i);continue;
          end;

          TC2 := TFALFormula(Owner.ListR.Items[i]);

          if TX.TrivialCheck(TC2) then
          begin
            truthstr := '';
            if Owner.SRcheck1 then
            begin
                  truthstr := 'DCF: ';
                  TX.Print(truthstr);
                  truthstr := truthstr + '//identical to R' + IntToStr(i)+';';
            end;
            if TC2.Degree < TX.Degree then
              begin
                TC2.Degree := TX.Degree;
                truthstr := truthstr + 'degree of R'+IntToStr(i)+' raised to: ' + FloatToStr(TX.Degree);
              end;
                if truthstr<>'' then Owner.OutputLines.Add(truthstr);

              Free; inc(nummol);
              Result := nil; Exit;
          end;
        end;
        inc(i);
      end;

    i := 0;

    if (TX.Cont.Ev = 0)and(Owner.RedCheck1 = false) then
    begin

      while i < Owner.ListR.Count do
      begin
        if (Owner.ListR.Items[i] <> nil) then
        begin
          if (TFALFormula(Owner.ListR.Items[i])).DisabledDCF then
          begin
            inc(i);continue;
          end;

          TC2 := TFALFormula(Owner.ListR.Items[i]);

          TC1 := TC2.CreateVF(TX, TImp);
          TC1.Cont.neg := true;
          TC1.Cont.EvalPol;

          TC3 := TC1.OptimizeX;

          TC1.DestroyVF(TC2, TX);

          if (TC3.Cont.Ev = -1) then
          begin
              truthstr := '';
              if Owner.SRcheck1 then
              begin
                truthstr := 'DCF: ';
                TX.Print(truthstr);
                truthstr := truthstr + '//consequent to R' + IntToStr(i)+':';
                TC2.Print(truthstr);
              end;
              if TC2.Degree < TX.Degree then
              begin
                TC3.Free;
                inc(i);
                continue;
                //TC2.Degree := TX.Degree;
                //truthstr := truthstr + 'degree of R'+IntToStr(i)+' raised to: ' + FloatToStr(TX.Degree);
              end;
                if truthstr<>'' then Owner.OutputLines.Add(truthstr);
              TC3.Free;
              Free; inc(nummol);
              Result := nil; Exit;
          end;

        end;
        TC3.Destroy;
        inc(i);
        Application.ProcessMessages;
        if Owner.stopf then Exit;
      end;

    end;

    Result := TX;

  if UsedMem > MaxMem then MaxMem:=UsedMem;
end;


{ Substitutes formula with its parent formula as a result of simpl. }
function TSub.SubstTo;
var anc : TSub;
begin
   anc := TSub(self.Ac);
   self.Ac := anc.Ac;
   if self.Ac is TSub then   begin
     if TSub(self.Ac).L = Anc then TSub(self.Ac).L := self
     else TSub(self.Ac).R := self;
   end
   else begin TFALFormula(self.Ac).Cont := self; end;
   if anc.L = self then anc.L := nil else anc.R := nil;
   if anc.neg then
     begin
       self.negate;
     end;
   anc.RFree;
   Result := self;
end;

procedure TFALFormula.Add;
var NewF:TSub;
begin
  if Cont<>nil then
  begin
    NewF:=T.Create;inc(numsub);
    NewF.Ac:=self;
    NewF.R:=X.Cont;
    NewF.L:=Cont;
    NewF.L.Ac:=NewF;
    NewF.R.Ac:=NewF;
    NewF.Ac:=self;
    Cont:=NewF;
    X.Cont:=nil;
    if ML<>nil then ML.R:=X.MF;
    if X.MF<>nil then X.MF.L:=ML;
    if X.ML<>nil then ML:=X.ML;
    if MF=nil then MF:=X.MF;
    X.Free;
  end
  else
  begin
    Cont:=X.Cont;
    X.Cont:=nil;
    Cont.Ac:=self;
    MF:=X.MF;
    ML:=X.ML;
    Parent1:=X.Parent1;
    Parent2:=X.Parent2;
    X.Free;
  end;
end;

procedure TPLProgram.ClearFormulas;
var i : longint;
begin
  for i := 0 to ListF.Count -1 do
    if ListF.Items[i] <> nil then TFALFormula(ListF.Items[i]).Free;
  ListF.Clear;
end;

procedure TPLProgram.ClearResolvents;
var i : longint;
begin
  for i := 0 to ListR.Count -1 do
    if ListR.Items[i] <> nil then TFALFormula(ListR.Items[i]).Free;
  ListR.Clear;
end;


destructor TPLProgram.Destroy;
begin
  ClearFormulas;
  ListF.Free;
  ListR.Free;
  MGun1.Free;
  inherited;
end;

destructor TFALFormula.Destroy;
begin
  if Cont <> nil then TSub(Cont).RFree; dec(numfl);
  Cont := nil;
  inherited;
end;

procedure TSub.RFree;
begin
  if L <> nil then TSub(L).RFree; L := nil;
  if R <> nil then TSub(R).RFree; R := nil;
  if (Q <> nil) then
  begin
    if (Ac is TSub) then
    begin
      if ((TSub(Ac).Pol = Pol) = TSub(Ac).Neg) then
        Q.Swap;
      TSub(Ac).AppendToQuant(Q);
    end;
    if (Ac is TFALFormula) then
      if (self <> TFALFormula(Ac).Cont)
      then
      begin
        if ((TFALFormula(Ac).Cont.Pol = Pol) = TFALFormula(Ac).Cont.Neg) then
        Q.Swap;
        TFALFormula(Ac).Cont.AppendToQuant(Q);
      end
      else Q.RFree;
    Q := nil;
  end;
  Free;
end;


procedure TAtom.RFree;
begin
  if Id <> nil then Id.RFree;
  Id := nil;
  if (Q <> nil) then
  begin
    if (Ac is TSub) then
    begin
      if ((TSub(Ac).Pol = Pol) = TSub(Ac).Neg) then
        Q.Swap;
      TSub(Ac).AppendToQuant(Q);
    end;
    if (Ac is TFALFormula) then
      if (self <> TFALFormula(Ac).Cont)
      then
      begin
        if ((TFALFormula(Ac).Cont.Pol = Pol) = TFALFormula(Ac).Cont.Neg) then
        Q.Swap;
        TFALFormula(Ac).Cont.AppendToQuant(Q);
      end
      else Q.RFree;
    Q := nil;
  end;
  if L <> nil then L.R := R;
  if R <> nil then R.L := L;
  if self = FirstMol then FirstMol := TAtom(FirstMol.R);
  if self = LastMol then LastMol := TAtom(LastMol.L);
  Free;
end;

procedure TQuant.Swap;
var i:integer;
begin
if slu then
begin
  for i := 0 to Count - 1 do
    if Items[i] <> nil then
    begin
      if TVariable(Items[i]).Ex = cforall then TVariable(Items[i]).Ex := cexists
      else TVariable(Items[i]).Ex := cforall;
      TVariable(Items[i]).Source := false;
    end;
end;
end;

procedure TQuant.RFree;
var i : integer; x : TVariable;
begin
  for i := 0 to Count-1 do
    if Items[i] <> nil then
    begin
      X := TVariable(Items[i]);
      if (X.id <> nil)
        then x.Id.RFree;
      X.Id := nil;
    end;
  for i := 0 to Count-1 do
    if Items[i] <> nil then
    begin
      X := TVariable(Items[i]);
      x.Free;
      Items[i] := nil;
    end;
  Free;
end;

procedure TVariable.RFree;
begin

end;

procedure TFunctor.RFree;
var i : integer; 
begin
  if Params <> nil then
  begin
    for i := 0 to Params.Count-1 do
      if Params.Items[i] <> nil then
      begin
        TId(Params.Items[i]).RFree;
        Params.Items[i] := nil; 
      end;
  end; 
  Free;
end;

destructor TFunctor.Destroy;
begin
  Params.Free;
  inherited;
end;

destructor TVariable.Destroy;
begin
  if id <> nil then
  begin
    id.RFree; 
    id := nil;
  end;
  inherited;
end;

procedure TId.RFree; 
begin
  Free;
end; 

constructor TFALFormula.Create(ow : TPLProgram);
begin
  Owner := ow; inc(numfl); Degree := 1;
end;

procedure TFALFormula.Print;
var i : longint; XM : TAtom; s : string;
begin
  if Symbolic<>'' then s1 := s1 + Symbolic else TSub(Cont).RPrint(s1);
  s1 := s1+':'+FloatToStr(Degree)+'; ';
end;

procedure TFALFormula.PrintRefutation;
begin
  truthstr := '';
  if Owner.Sources then
    truthstr := '['+rsv1+'&'+rsv2+']'+' : ';
  truthstr := truthstr+' YES (refutation deg.):'+FloatToStr(FN1.Degree)+cFmlSep;
  Owner.OutputLines.Add(truthstr);
  FN1.PrintVars; Owner.Proved := true;
  if BestDegree < FN1.Degree then
  begin
    BestDegree := FN1.Degree;
    BestSubst := FN1.PrintVarsToStr;
  end;
  if Owner.Oneref1 then retf := true;
end;

procedure TQuant.RPrint;
var i : integer; LastV : TVariable;
begin
  if CurProg.Quant1 then
  begin
    LastV := nil;
    for i := 0 to Count-1 do
      begin
        if (Items[i] <> nil)
          and(TVariable(Items[i]).Id = nil)
          then
        begin
          if (i > 0)and(LastV <> nil)and(LastV.Ex = TVariable(Items[i]).Ex)
          then st := st+', ' else st := st+TVariable(Items[i]).Ex;
          if not(TVariable(Items[i]).Source) then st := st+'!';
          st := st+TVariable(Items[i]).Name;
          if TVariable(Items[i]).SubId > 0 then st := st +'#' + IntToStr(TVariable(Items[i]).SubId);
          LastV := TVariable(Items[i]);
        end;
      end;
  end;
end;

procedure TSub.RPrint; 
begin
  if Ev <> 0 then
  begin
    if Ev = 1 then st := st+' '+csup else st := st+' '+cinf;
    Exit; 
  end;
  if Q <> nil then Q.RPrint(st);
  if neg then st := st+cneg;
  if Ac is TFALFormula then
  begin
    if (neg)or(Q <> nil) then st := st+'[';
    if L <> nil then TSub(L).RPrint(st);
    st := st+GetSymbol;
    if R <> nil then TSub(R).RPrint(st);
    if (neg)or(Q <> nil) then st := st+']';
  end
  else
  begin
    if (neg)or(TSub(Ac).GetP > GetP)or
       ((TSub(Ac) is TImp)and(Self is TImp)and(Self = TSub(Ac).R))
       then
    begin
      st := st+'[';
      if L <> nil then TSub(L).RPrint(st);
      st := st+GetSymbol; 
      if R <> nil then TSub(R).RPrint(st);
      st := st+']'; 
    end
    else
    begin
      if (Q <> nil) then st := st+'[';
      if L <> nil then TSub(L).RPrint(st);
      st := st+GetSymbol;
      if R <> nil then TSub(R).RPrint(st);
      if (Q <> nil) then
      begin
        st := st+']';
      end;
    end;
  end;
end;


function TImp.GetSymbol : string;
begin
  if Ev = 0 then GetSymbol := cImp;
end;

function TDis.GetSymbol : string;
begin
  if Ev = 0 then GetSymbol := cor;
end;
function TCon.GetSymbol : string;
begin
  if Ev = 0 then GetSymbol := cand;
end;

function TImp.GetP : integer;
begin
  GetP := 2;
end;

function TDis.GetP : integer;
begin
  GetP := 3;
end;

function TCon.GetP : integer;
begin
  GetP := 4;
end;

function TAtom.GetP : integer;
begin
  GetP := 5;
end;

procedure TAtom.RPrint;
begin
  if Ev <> 0 then
  begin
    if Ev = 1 then st := st+' '+csup else st := st+' '+cinf;
    Exit;
  end;
  if Q <> nil then
  begin
    Q.RPrint(st);
    st := st+' ';
  end;
  if neg then st := st+cneg;
  if Id <> nil then Id.RPrint(st);
end;

procedure TInteger.RPrint; 
begin
  st := st+IntToStr(Id);
end; 

procedure TReal.RPrint; 
begin
  st := st+FloatToStr(Id);
end;

procedure TVar.RPrint;
begin
  st := st+Id;
end;

procedure TVariable.RPrint;
begin
  if Id <> nil then Id.RPrint(st) else
  begin
    st := st+Name;
    if SubId > 0 then st := st +'#' + IntToStr(SubId);
  end;
end;

procedure TVariable.PrintAs;
var st1 : string;
begin
    st1 := '';
    if Id <> nil then Id.RPrint(st1);
    if not(st1[1] in upper) then
    begin
      if st <> '' then st := st+', ';
      st := st+Name+' = '+st1;
    end;
end;

constructor TInteger.Create;
begin
  try
     Id := StrToInt(i); 
     except
     on EConvertError do Error := InvRealVal;
     end;
end; 

constructor TInteger.CreateN;
begin
  id := i;
end;

constructor TReal.CreateN;
begin
  id := i;
end;

constructor TReal.Create;
begin
   try
     Id := StrToFloat(i);
     except
     on EConvertError do Error := InvRealVal;
     end;
end;

constructor TVariable.Create;
begin
  Id := nil;
  NewId := nil;
  Name := i.Id;
  New := nil;
  Source := false;
  IsInMGU := false;
  identifier := varident;
  inc(varident);
end;

constructor TVariable.CreateV;
begin
  Id := nil;
  NewId := nil;
  Ex := i.Ex;
  Name := i.Name;
  Ap := i.Ap;
  New := nil;
  Watch := i.Watch;
  Source := i.Source;
  IsInMGU := false;
  identifier := i.identifier;
end;

constructor TVariable.CreateS;
begin
  Id := nil;
  NewId := nil;
  Name := i;
  New := nil;
  IsInMGU := false;
end;

constructor TVar.Create;
begin
  Id := i;
end;

constructor TFunctor.Create;
begin
  Functor := i;
  Params := TList.Create; 
end; 

constructor TFunctorA.Create;
begin
  Functor := i;
end;

procedure TFunctor.RPrint;
var i : integer;
begin
  if Functor[1] in InfixPred then
  begin
    if Params <> nil then
    begin
      for i := 0 to Params.Count-1 do
      begin
        if Params.Items[i] <> nil then
        begin
          TId(Params.Items[i]).RPrint(st); 
          if i < Params.Count-1 then st := st+Functor;
        end;
      end;
    end;
  end
  else
  if Functor[1] in InfixOper then
  begin
    if Params <> nil then
    begin
      if Params.Count = 1 then
      begin
        st := st+Functor;
        TId(Params.Items[0]).RPrint(st);
      end
      else
      for i := 0 to Params.Count-1 do
      begin
        if Params.Items[i] <> nil then
        begin
          if (TObject(Params.Items[i]) is TFunctor)and
             (Functor[1] in ['*', '/'])
              and(TFunctor(Params.Items[i]).Functor[1] in ['+', '-'])
          then
          begin
            st := st+'(';
            TId(Params.Items[i]).RPrint(st); 
            st := st+')';
          end
          else TId(Params.Items[i]).RPrint(st);
          if i < Params.Count-1 then st := st+Functor;
        end;
      end;
    end;
  end
  else
  begin
    st := st+Functor;
    if Params <> nil then
    begin
      if (Functor <> '(')and(Functor <> '{') then st := st+'(';
      for i := 0 to Params.Count-1 do
      begin
        if Params.Items[i] <> nil then
        begin
          TId(Params.Items[i]).RPrint(st);
          if i < Params.Count-1 then st := st+', ';
        end;
      end;
      if (Functor <> '{') then st := st+')' else st := st+'}'; 
    end;
  end;
end;

procedure TFunctorA.RPrint;
begin
  st := st+Functor;
  if Functor[1] = '{' then st := st+'}'
  else if Functor[1] = '(' then st := st+')';
end;

procedure TSub.Negate;
var i : integer;
begin
  neg := not(neg);
  if Q<>nil then
  begin
    for i := 0 to Q.Count-1 do
      if Q.Items[i] <> nil then
        case TVariable(Q.Items[i]).Ex of
          cexists : TVariable(Q.Items[i]).Ex := cforall;
          cforall : TVariable(Q.Items[i]).Ex := cexists;
        end;
  end;
  EvalPol;
end;

procedure TSub.OptQ;
var i,j,num : integer;
begin
  for i := 0 to Q.Count-1 do if Q.Items[i]<>nil then TVariable(Q.Items[i]).SubId := 0;
  for i := 0 to Q.Count-1 do
  begin
    if (TVariable(Q.Items[i]).Id = nil) then num := 1
    else num:=0;
    if (TVariable(Q.Items[i]).SubId = 0)and(TVariable(Q.Items[i]).Id = nil) then
    for j := 0 to Q.Count-1 do
    begin
      if (i<>j)and(TVariable(Q.Items[i]).Name = TVariable(Q.Items[j]).Name)and
        (TVariable(Q.Items[j]).Id = nil) then
        begin
          TVariable(Q.Items[j]).SubId:=num;inc(num);
        end;
    end;
  end;
end;

procedure TSub.AppendToQuant;
var i : integer;
begin
  if (Q <> nil)and(Q.Count = 0) then
  begin
    Q.Free;
    Q := nil;
  end;
  if Q = nil then
  begin
    Q := Q1;
  end
  else
  begin
    for i := 0 to Q.Count-1 do
      if (Q.Items[i] <> nil)
        then
        begin
          TVariable(Q.Items[i]).Ap := true;
          Q1.Add(Q.Items[i]);
        end;
    Q.Free;
    Q := Q1;
  end;
  if (Q1 <> nil)and(Q1.Count = 0) then
  begin
    Q1.Free;
    Q1 := nil;
  end;
  if Q <> nil then
  for i := 0 to Q.Count-1 do
      if (Q.Items[i] <> nil)
        then
        begin
          TVariable(Q.Items[i]).Master := self;
        end;
  OptQ;
end;

function TFunc.Compare(X : TId) : Integer;
var i : integer;
begin
  i := 0;
  if X is TFunc then
  begin
    if self is TFunctor then
    begin
      if X is TFunctor then
      begin
        Result := CompareText(TFunctor(self).Functor, TFunctor(X).Functor);
        if Result = 0 then
        begin
          if(TFunctor(self).Params = nil) then Result := -1
          else if (TFunctor(X).Params = nil) then
            Result := 0
          else
          if TFunctor(self).Params.Count > TFunctor(X).Params.Count then
            Result := 1
          else if (TFunctor(self).Params.Count < TFunctor(X).Params.Count) then
            Result := -1
          else
          while (i < TFunctor(self).Params.Count)and(result = 0) do
          begin
            if (TFunctor(self).Params.Items[i] <> nil)and
            (TFunctor(X).Params.items[i] <> nil) then
            Result := (TId(TFunctor(self).Params.Items[i]).
                   Compare(TId(TFunctor(X).Params.Items[i])));
            inc(i);
          end;
        end;
      end
      else
      begin
        if TFunctor(self).Params <> nil then Result := 1 else
        Result := CompareText(TFunctor(self).Functor, TFunctora(X).Functor);
      end;
    end
    else
    begin
      if X is TFunctor then
      begin
        if TFunctor(X).Params <> nil then Result := -1 else
        Result := CompareText(TFunctora(self).Functor, TFunctor(X).Functor);
      end
      else
      begin
        Result := CompareText(TFunctor(self).Functor, TFunctora(X).Functor);
      end;
    end;
  end
  else if X is TVariable then
  begin
    if TVariable(X).Id <> nil then Result := Compare(TVariable(X).Id)
    else Result := 1;
  end
  else Result := 1;
end;

function TVar.Compare(X : TId) : integer; 
begin
    if X is TVar then Result := CompareText(Id, TVar(X).Id)
    else if X is TVariable then Result := Compare(TVariable(X).Id)
    else if X is TFunc then Result := -1 else Result := 1;
end;

function CompN(I1, I2 : Extended) : Integer;
begin
  if I1 > I2 then Result := 1 else if I1 < I2 then Result := -1
  else Result := 0; 
end;

function TReal.Compare(X : TId) : integer; 
begin
   if (X is TReal) then
     Result := CompN(Id, TReal(X).Id)
   else if (X is TInteger) then
     Result := CompN(Id, TInteger(X).Id)
   else if (X is TVariable)and(TVariable(X).Id <> nil)
        then Result := Compare(TVariable(X).Id)
   else Result := -1;
end;

function TInteger.Compare(X : TId) : integer;
begin
   if (X is TReal) then
     Result := CompN(Id, TReal(X).Id)
   else if (X is TInteger) then
     Result := CompN(Id, TInteger(X).Id)
   else if (X is TVariable)and(TVariable(X).Id <> nil)
        then Result := Compare(TVariable(X).Id)
   else Result := -1;
end;


function TVariable.Compare(X : TId) : integer;
begin
  if Id <> nil then Result := Id.Compare(X)
  else
  begin
    if (X is TVariable) then
      if (TVariable(X).Id <> nil) then
        if  TVariable(X).Id is TVar then
          Result := CompareText(Name, TVar(TVariable(X).Id).Id)
        else if (TVariable(X).Id is TNumber)
             then Result := 1
             else Result := -1
      else Result := CompareText(Name, TVariable(X).Name)
    else if (X is TNumber)
              then Result := 1
         else Result := -1;
  end;
end;

function TReal.GetNumber;
begin
  Result := Id; 
end;

function TInteger.GetNumber;
begin
  Result := Id;
end;

var LastSub : TSub;

{ Procedure for assignment of complex variable object for every occurence of
  the variable. }
function VarReAssign(o1 : TObject; var tp : char; argums : array of const)
                                                                 : TObject;
var S1 : TSub; i : integer; O : TVariable;
begin
  Result := o1;
  if o1 is TSub then
  begin
    LastSub := TSub(o1);
    if LastSub.Q <> nil then
    with LastSub.Q do
    begin
      for i := 0 to Count-1 do
      begin
        TVariable(Items[i]).Master := LastSub; { Resets variable info. }
        TVariable(Items[i]).Ap := false;
      end;
    end;
    if o1 is TAtom then tp := '3';
  end
  else if (o1 is TVar) then
  begin
  S1 := LastSub; O := nil;
  if S1.Q <> nil then
    with S1.Q do
    begin
      for i := 0 to Count-1 do      { Searches for right TVariable object. }
        if TVariable(Items[i]).Name = TVar(o1).Id then O := TVariable(Items[i]);
    end;
    while not(S1.Ac is TFALFormula)and(O = nil) do
    begin
      S1 := TSub(S1.Ac);
      if S1.Q <> nil then
      with S1.Q do
      begin
        for i := 0 to Count-1 do
          if TVariable(Items[i]).Name = TVar(o1).Id then
                                      O := TVariable(Items[i]);
      end;
    end;
    if O <> nil then
    begin
      o1.Free;
    end
    else
    begin
      if S1.Q = nil then S1.Q := TQuant.Create;
      O := TVariable.Create(TVar(o1));
      O.Ap := false;
      O.Master := S1;    { If not found, creates new variable on  }
      o1.Free;           {  the top level.                        }
      O.Ex := cforall;
      O.Source := true;
      S1.Q.Insert(0,O);
    end;
    Result := O;
  end;
end;

procedure TFALFormula.PostComp;
var TT1 : TAtom;
begin
  LastSub := nil;
  Dive('4', VarReAssign, [nil]); { Assignment of variables. }
  Dive('3', EvalBuilt, [nil]);  { Evaluation of infix operators. }
  Cont.EvalPol;
  if not (self is TQuery) then MakeSymbolic;
end;

procedure TSub.EvalPol;
begin
  if Ac is TFALFormula then if Neg then Pol := -1 else Pol := 1
  else
  begin
    if (Ac is TImp)and(TSub(Ac).L = self) then Pol := -TSub(Ac).Pol
    else Pol := TSub(Ac).Pol;
    if Neg then Pol := -Pol;
  end;
  if (self is TAtom) then Exit
  else { Equivalence causes zero priority. }
  begin
    if L <> nil then L.EvalPol;
    if R <> nil then R.EvalPol;
  end;
end;


procedure TQuery.NegQ;
var M : TAtom;
begin
  Cont.Negate;   { Negates tree and negates polarity. }
  Cont.EvalPol;
  MakeSymbolic;
end;

var
  SCompare : TListSortCompare;

procedure QuickSort(SortList : PPointerList; L, R : Integer);
var
  I, J : Integer;
  P, T : Pointer;
begin
  repeat
    I := L;
    J := R;
    P := SortList^[(L + R) shr 1];
    repeat
      while (I < R)and
      (SCompare(SortList^[I], P) < 0)
        do Inc(I);
      while (J > L)and
      (SCompare(SortList^[J], P) > 0)
        do Dec(J);
      if I <= J then
      begin
        T := SortList^[I];
        SortList^[I] := SortList^[J];
        SortList^[J] := T;
        Inc(I);
        Dec(J);
      end;
    until I > J;
    if L < J then QuickSort(SortList, L, J);
    L := I;
  until I >= R;
end;

{ Prints variables' assignment. }
procedure TFALFormula.PrintVars;
var tempstr : string;
begin
      tempstr := PrintVarsToStr;
      if tempstr <>'' then Owner.OutputLines.Add(tempstr);
end;

function TFALFormula.PrintVarsToStr:string;
var i : longint; s : string; Qt : TQuant; CV, LastVar : TVariable;
begin
  s := ''; LastVar := nil;
  if (Cont.Ev <> 0)and(Cont.Q <> nil) then
  begin
    Qt := Cont.Q;
    i := 0;
    while i < Qt.Count do
    begin
     CV := TVariable(Qt.Items[i]);
     if (Qt.Items[i] <> nil)and(CV.Id <> nil) then
     begin
        begin
          if (CV.Id <> nil)and(CV.Watch <> nil) then
          begin
            CV.PrintAs(s);
            LastVar := CV;
          end;
          inc(i);
        end;
      while (i < Qt.Count)and(LastVar <> nil)and
         ((LastVar.Watch = TVariable(Qt.Items[i]).Watch))
         and(((LastVar.Id <> nil)and(TVariable(Qt.Items[i]).Id <> nil)
         and(LastVar.Id.Compare(TVariable(Qt.Items[i]).Id) = 0))or
         (TVariable(Qt.Items[i]).Id = nil)) do inc(i);
     end  {  Prints vars uniquely and only assigned. }
     else inc(i);
    end;
    if s <> '' then
    begin
      s := s+cFmlSep;
    end;
  end;
  PrintVarsToStr := s;
end;


procedure TQuery.EvalQuery;
begin
  NegQ;
  Owner.ListF.Add(self);
  Support := true;
  Dive('v', nil, [nil]); { Marks variables requested by a query. }
  Owner.Consist;
  Owner.ListF.Remove(self);
  Free;
end;

var OGP : TObject;

{ Now follow the implementations of the Dive procedure for several
  types of recursive browse of the syntactical tree.
  '1' - preorder, '2' - inorder, '3' - postorder,
  '4' - postorder only without Q nodes, 'r' - copy tree,
  'v' - marks variables watch property, 'q' - makes the new
    variable assignment, 'd' - deletes all temporal changes
    in a variable, 's' - simplification process. }
function TAtom.Dive;
var i : longint;
begin
  case tp of
  '1' :
    begin
      Proc(self, tp, argums);
      if Q <> nil then Q.Dive(tp, proc, argums);
      if Id <> nil then  Id.Dive(tp, proc, argums);
    end;
  '2', '3' :
    begin
      if Q <> nil then Q := Q.Dive(tp, proc, argums);
      if Id <> nil then  Id := Id.Dive(tp, proc, argums);
      Result := TSub(Proc(self, tp, argums));
    end;
  '4' :
    begin
      Result := TSub(Proc(self, tp, argums));
      if Id <> nil then  Id := Id.Dive(tp, proc, argums);
    end;
    'r' :
    begin
      Result := TAtom(ClassType.Create);
      Result.Neg := Neg;

      if invert then
      begin
        Result.Ev := -Ev;
        TempEv := 0;
      end
      else Result.Ev := Ev;

      TAtom(Result).Pol := Pol;
      Result.Ac := Ancs;
      if LastMol <> nil then
      begin
        Result.L := LastMol;
        LastMol.R := Result;
        LastMol := TAtom(Result);
      end
      else
      begin
        FirstMol := TAtom(Result);
        LastMol := TAtom(Result);
      end;

      if Q <> nil then
      begin
        TAtom(Result).Q := Q.Dive(tp, proc, argums);
        if Result.Q <> nil then
        with Result.Q do
        for i := 0 to Count-1 do
          TVariable(Items[i]).Master := Result;
      end;
      if Id <> nil then  TAtom(Result).Id := Id.Dive(tp, proc, argums);
      if Q <> nil then Q.Dive('q', proc, argums);
    end;
    'v' :
    begin
      if Q <> nil then Q.Dive(tp, proc, argums);
    end;
    'd' :
    begin
      Ev := 0;
      if Q <> nil then
      begin
        Q.Dive(tp, proc, argums);
        if Q.Count = 0 then
        begin
          Q.Free;
          Q := nil;
        end;
      end;
    end;
  end;
end;



function TSub.Dive;
var i : longint;
begin
  case tp of
  '1' :
    begin
      Proc(self, tp, argums);
      if Q <> nil then  Q.Dive(tp, proc, argums);
      if L <> nil then  L.Dive(tp, proc, argums);
      if R <> nil then  R.Dive(tp, proc, argums);
    end;
  '2' : 
    begin
      if Q <> nil then  Q.Dive(tp, proc, argums);
      if L <> nil then  L.Dive(tp, proc, argums);
      Proc(self, tp, argums);
      if R <> nil then  R.Dive(tp, proc, argums);
    end;
  '3' :
    begin
      if Q <> nil then  Q := Q.Dive(tp, proc, argums);
      if L <> nil then  L := L.Dive(tp, proc, argums);
      if R <> nil then  R := R.Dive(tp, proc, argums);
      Result := TSub(Proc(self, tp, argums)); 
    end;
  '4' :
    begin
      Result := TSub(Proc(self, tp, argums));
      if L <> nil then  L := L.Dive(tp, proc, argums);
      if R <> nil then  R := R.Dive(tp, proc, argums); 
    end;
    'r' :
    begin
      Result := TSub(ClassType.Create);
      Result.Neg := Neg; Result.Ev := Ev;
      Result.Pol := Pol;
      Result.Ac := Ancs; LastSub := Result;
      if Q <> nil then
      begin
        Result.Q := Q.Dive(tp, proc, argums); 
        if Result.Q <> nil then
        with Result.Q do
        for i := 0 to Count-1 do
          TVariable(Items[i]).Master := Result;
      end;
      Ancs := Result; LastSub := Result;
      if L <> nil then Result.L := L.Dive(tp, proc, argums);
      Ancs := Result; LastSub := Result;
      if R <> nil then Result.R := R.Dive(tp, proc, argums);
      LastSub := Result;
      if Q <> nil then Q.Dive('q', proc, argums);
    end;
    'd' :
    begin
      if L <> nil then  L.Dive(tp, proc, argums);
      if R <> nil then  R.Dive(tp, proc, argums);
      if Q <> nil then
      begin
        Q.Dive(tp, proc, argums);
        if Q.Count = 0 then
        begin
          Q.Free;
          Q := nil;
        end;
      end;
    end;
    's' :
    begin
      if (L <> nil)and(not(L is TAtom))and((L.Ev = 0))
        then  L := L.Dive(tp, proc, argums);
      if (R <> nil)and(not(R is TAtom))and((R.Ev = 0))
        then  R := R.Dive(tp, proc, argums);
      Result := TSub(Proc(self, tp, argums));
    end;
    'v' :
    begin
      if Q <> nil then Q.Dive(tp, proc, argums);
    end; 
  end;
end;

var XGP2 : TObject;

function TQuant.Dive;
var i : longint; X : TVariable;
begin
  case tp of
  '1' :
    begin
      Proc(self, tp, argums);
      for i := 0 to Count-1 do
      if Items[i] <> nil then
      begin
        TVariable(Items[i]).Dive(tp, proc, argums); 
      end;
    end;
  '2', '3' :
    begin
      for i := 0 to Count-1 do
      if Items[i] <> nil then
      begin
        Items[i] := TVariable(Items[i]).Dive(tp, proc, argums);
      end;
      Result := TQuant(Proc(self, tp, argums));
    end;
    'r' :
    begin
      Result := TQuant(ClassType.Create);
      for i := 0 to Count-1 do
      if Items[i] <> nil then
      begin
        X := TVariable(Items[i]);
        X.Used := true;
        if not(X.UnSub) then
        begin
          X.New := TVariable.CreateV(TVariable(Items[i]));
          Result.Add(X.New);
        end;
        if X.IsInMGU
        then X.New.Source := false;
      end;
      if Result.Count = 0 then
      begin
        Result.Free;
        Result := nil;
      end;
    end;
    'q' :
     begin
       for i := 0 to Count-1 do
      if Items[i] <> nil then
      begin
        X := TVariable(Items[i]);
        if (X.NewId <> nil)
          and(X.New.Id = nil) then
            X.New.Id := X.NewId.Dive('p', proc, argums)
        else
        if (X.Id <> nil)
         and(X.New.Id = nil) then
         X.New.Id := X.Id.Dive('r', proc, argums);
      end;
     end;
     'v' :
     begin
       for i := 0 to Count-1 do
        begin
          if Items[i] <> nil then
          begin
            X := TVariable(Items[i]);
            X.New := nil;
            X.Id := nil;
            X.Used := false;
            X.Watch := X;
          end;
        end;
     end;
     'd' :
     begin
       for i := 0 to Count-1 do
        begin
          if Items[i] <> nil then
          begin
            if (TVariable(Items[i]).Used = false)
             and(not((TVariable(Items[i]).Watch <> nil)
             and(TVariable(Items[i]).Id <> nil))) then
            begin
              dest := true;
              TVariable(Items[i]).Destroy;
              Items[i] := nil;
            end
            else
            begin
              TVariable(Items[i]).Unsub := false;
              TVariable(Items[i]).Used := false;
              TVariable(Items[i]).New := nil;
              TVariable(Items[i]).NewId := nil;
              TVariable(Items[i]).IsInMgu := false;
            end;
          end;
        end;
       Pack;
     end;
  end;
end;

var unres, unres1, ocres : shortint; FiV : TVariable;
    XGP1, XGPP : TObject;

{  '5' - unification, '6' - check for occurence of a variable.,
   'P' - assigns old variable used in a substitution,
   'P' - assigns new variable used in a substitution. }
function TId.Dive;
var i : longint; X : TVariable;
begin

  case tp of
  '1' :
    begin
      Proc(self, tp, argums);
      if self is TFunctor then
        for i := 0 to TFunctor(self).Params.Count-1 do
        if TFunctor(self).Params.Items[i] <> nil then
        begin
          TId(TFunctor(self).Params.Items[i]).Dive(tp, proc, argums);
        end;
    end; 
  '2' :
    begin
      if (self is TFunctor) then
        if  TFunctor(self).Functor[1] in infixpred then
        begin
         for i := 0 to TFunctor(self).Params.Count-1 do
         if TFunctor(self).Params.Items[i] <> nil then
         begin
          TId(TFunctor(self).Params.Items[i]).Dive(tp, proc, argums);
          if (i = 0) then Proc(self, tp, argums);
         end;
        end
        else
        begin
           Proc(self, tp, argums);
           for i := 0 to TFunctor(self).Params.Count-1 do
           if TFunctor(self).Params.Items[i] <> nil then
           begin
            TId(TFunctor(self).Params.Items[i]).Dive(tp, proc, argums);
           end;
        end
      else Proc(self, tp, argums);
    end;
  '3', '4' : 
    begin
      if self is TFunctor then
      begin
        for i := 0 to TFunctor(self).Params.Count-1 do
        if TFunctor(self).Params.Items[i] <> nil then
        begin
          TFunctor(self).Params.Items[i] :=
            TId(TFunctor(self).Params.Items[i]).Dive(tp, proc, argums); 
        end;
      end
      else if self is TVariable then
      begin
        if TVariable(self).Id <> nil
          then TVariable(self).Id := TVariable(self).Id.Dive(tp, proc, argums)
      end;
      Result := TId(Proc(self, tp, argums));
    end;
      '5' :
    begin
      XGP1 := self;
      Result := self;
      if argums[0].VObject = XGP1 then
        Exit;
      if argums[0].VObject is TVariable then
        argums[0].VObject := TVariable(argums[0].VObject).CompInMgu;
      if argums[0].VObject = XGP1 then
        Exit;
      if self is TVariable then
        XGP1 := TVariable(self).CompInMgu;
      if argums[0].VObject = XGP1 then
        Exit;
      while (argums[0].VObject is TVariable)and
        (TVariable(argums[0].VObject).Id <> nil) do
      begin
           argums[0].VObject := TVariable(argums[0].VObject).Id;
           if argums[0].VObject = XGP1 then
             Exit;
           if argums[0].VObject is TVariable then
           argums[0].VObject := TVariable(argums[0].VObject).CompInMgu;
      end;
      while (XGP1 is TVariable)and(TVariable(XGP1).Id <> nil) do
      begin
           XGP1 := TVariable(XGP1).Id;
           if argums[0].VObject = XGP1 then
             Exit;
           if argums[0].VObject is TVariable then
           argums[0].VObject := TVariable(argums[0].VObject).CompInMgu;
      end;
      if argums[0].VObject is TVariable then
        if TVariable(argums[0].VObject).NewId <> nil then
           argums[0].VObject := TVariable(argums[0].VObject).NewId;
      if argums[0].VObject = XGP1 then
        Exit;
      if XGP1 is TVariable then
         if TVariable(XGP1).NewId <> nil then
           XGP1 := TVariable(XGP1).NewId;
      if argums[0].VObject = XGP1 then
        Exit;
      { Searches for possible assignments of a variable. }
      Result := TId(XGP1);
      if XGP1 <> self then TId(XGP1).Dive(tp, proc, argums)
      else
      begin
      Proc(self, tp, argums);
      if (self is TFunctor)and(unres > 0) then
       begin
        i := 0;
        while (i < TFunctor(self).Params.Count)and(unres <> 0) do
        if (TFunctor(self).Params.Items[i] <> nil)
         and(TFunctor(argums[0].VObject).Params.Items[i] <> nil) then
        begin
          TId(TFunctor(self).Params.Items[i]).Dive(tp, proc,
          [TObject(TFunctor(argums[0].VObject).Params.Items[i])]);
          inc(i);
        end
        else inc(i);
       end;
      end;
    end;
    '6' :
    begin
      Proc(self, tp, argums);
      if self is TFunctor then
      begin
        i := 0;
        while (i < TFunctor(self).Params.Count)
          and(ocres <> 0) do
        if TFunctor(self).Params.Items[i] <> nil then
        begin
          TId(TFunctor(self).Params.Items[i]).Dive(tp, proc, argums);
          inc(i);
        end
        else inc(i);
      end
      else if self is TVariable then
      begin
        X := TVariable(self);
        XGPP := X.CompInMgu;
        if self <> XGPP then TId(XGPP).Dive(tp, proc, argums)
        else
        if X.Id <> nil then X.Id.Dive(tp, proc, argums);
      end;
    end;
    'r' :
    begin
      if self is TFunctor then
      begin
        Result := TFunctor.Create(TFunctor(self).Functor);
        if TFunctor(self).Params <> nil then
        for i := 0 to TFunctor(self).Params.Count-1 do
        if TFunctor(self).Params.Items[i] <> nil then
        begin
          TFunctor(Result).Params.Add(
          TId(TFunctor(self).Params.Items[i]).Dive(tp, proc, argums));
        end;
      end
      else if self is TVar then
      begin
        Result := TVar.Create(TVar(self).Id);
      end
      else if self is TFunctora then
      begin
        Result := TFunctora.Create(TFunctora(self).Functor);
      end
      else if self is TReal then
      begin
        Result := TReal.CreateN(TReal(self).Id);
      end
      else if self is TInteger then
      begin
        Result := TInteger.CreateN(TInteger(self).Id);
      end
      else if self is TVariable then
      begin
         Result := TVariable(self).New;
      end;
    end;
    'p' :
    begin
      if self is TFunctor then
      begin
        Result := TFunctor.Create(TFunctor(self).Functor);
        if TFunctor(self).Params <> nil then
        for i := 0 to TFunctor(self).Params.Count-1 do
        if TFunctor(self).Params.Items[i] <> nil then
        begin
          TFunctor(Result).Params.Add(
          TId(TFunctor(self).Params.Items[i]).Dive(tp, proc, argums));
        end;
      end
      else if self is TFunctora then
      begin
        Result := TFunctora.Create(TFunctora(self).Functor);
      end
      else if self is TReal then
      begin
        Result := TReal.CreateN(TReal(self).Id);
      end
      else if self is TInteger then
      begin
        Result := TInteger.CreateN(TInteger(self).Id);
      end
      else if self is TVariable then
        begin
          X := TVariable(self);
          if X.New = nil then
          begin
            X.New := TVariable.CreateV(TVariable(self));
            X.Unsub := true;
            X.Used := true;
            newQ.Add(X.New);
            if X.Master.Pol < 0 then
              if X.New.Ex = cforall then X.New.Ex := cexists
              else X.New.Ex := cforall
              else X.New.Ex := cexists;

          end;
          Result := X.New;
        end;
    end;
    'P' :
    begin
      if self is TFunctor then
      begin
        Result := TFunctor.Create(TFunctor(self).Functor);
        if TFunctor(self).Params <> nil then
        for i := 0 to TFunctor(self).Params.Count-1 do
        if TFunctor(self).Params.Items[i] <> nil then
        begin
          TFunctor(Result).Params.Add(
          TId(TFunctor(self).Params.Items[i]).Dive(tp, proc, argums));
        end;
      end
      else if self is TFunctora then
      begin
        Result := TFunctora.Create(TFunctora(self).Functor);
      end
      else if self is TReal then
      begin
        Result := TReal.CreateN(TReal(self).Id);
      end
      else if self is TInteger then
      begin
        Result := TInteger.CreateN(TInteger(self).Id);
      end
      else if self is TVariable then
        begin
          X := TVariable(self);
          X.Unsub := true;
          X.Used := true;
          Result := X;
        end;
    end;
  end;
end;

function TFALFormula.Dive;
begin
  if Cont <> nil then Cont.Dive(tp, proc, argums);
  result := self;
end;

function TFALFormula.DiveD;
var F1 : TAtom;
begin
  F1 := MF;
  F1.Dive(tp, proc, argums);
  while F1 <> ML do
  begin
    F1.Dive(tp, proc, argums);
    F1 := TAtom(F1.R);
  end;
  Result := self;
end;

{ Checks, if all superior variables of an existencial variable are
  assigned. }
var ox:TObject;

function TVariable.MAssigned;
var i : longint; ASub : TSub;
begin
  Result := true;
  if ((abs(Master.Pol) = 1)and(Ex = cforall) = (Master.Pol >= 0) = (Master.Neg = false))
    then Exit;
  ASub := Master;
  if Master.Q <> nil then
  with Master.Q do
  begin
    for i := IndexOf(self)-1 downto 0 do
    begin
      if TVariable(Items[i]).Source then
      if (((abs(TVariable(Items[i]).Master.Pol) = 1)and((TVariable(Items[i]).Ex = cforall) =
       ((TVariable(Items[i]).Master.Pol >= 0) = (TVariable(Items[i]).Master.Neg = false)))))
       and(TVariable(Items[i]).Id = nil)
        and((TVariable(Items[i]).CompInMgu = TVariable(Items[i])))
        then begin Result := false; Exit; end;
    end;
  end;
  if ASub.Ac is TSub then ASub := TSub(ASub.Ac) else ASub := nil;
  while ASub <> nil do
  begin
    if ASub.Q <> nil then
    with ASub.Q do
    for i := Count-1 downto 0 do
     begin
       if TVariable(Items[i]).Source then
       if (((abs(TVariable(Items[i]).Master.Pol) = 1)and((TVariable(Items[i]).Ex = cforall) =
       ((TVariable(Items[i]).Master.Pol >= 0) = (TVariable(Items[i]).Master.Neg = false)))))
       and(TVariable(Items[i]).Id = nil)
          and((TVariable(Items[i]).CompInMgu = TVariable(Items[i])))
          then begin Result := false; Exit; end;
     end;
    if ASub.Ac is TSub then ASub := TSub(ASub.Ac) else ASub := nil;
  end;
end;

{ Decides, if o1 has greater priority than o2. }
function IsHigherObject(var o1, o2 : TObject) : boolean;
var swap : TObject;
begin
  Result := false;

  if (o1 is TVariable) then
  begin
    if ((abs(TVariable(o1).Master.Pol) = 1)and((TVariable(o1).Ex = cforall) =
       ((TVariable(o1).Master.Pol >= 0) = (TVariable(o1).Master.Neg = false))))
       then
    begin
      ox:=o1;
      if (o2 is TVariable)
        and(not(TVariable(o2).MAssigned)) then
        begin
          Result := false;
          Exit;
        end
      else begin
             Result := true;
             Exit;
           end;
    end
  end;
end;

{  Unifies two Id terms. }
function IdUnif(o : TObject; var tp : char; argums : array of const) : TObject;
var R : TObject;
begin
  Result := o;
  if (o is TId)and(argums[0].VType = vtObject)and(unres <> 0) then
  begin
    if (argums[0].VObject) is TVariable then
    begin
      begin
        if (o is TVariable)and(TVariable(o).identifier = TVariable(argums[0].VObject).identifier)
        then begin unres := -1; Exit; end;

        if (TId(o).OccurCheck(TVariable(argums[0].VObject)))and
          (not(CurProg.Quant1)
          or(IsHigherObject(argums[0].VObject, o)))
            then
        begin
          mgu1.Add(argums[0].VObject);
          mgu1.Add(o);
          unres := -1; Exit;
        end
        else begin if not(o is TVariable) then unres1 := 0; end;
      end
    end;
    if (o is TFunctor) then
    begin
     unres := 1;
     if not((argums[0].VObject is TFunctor)and
       (TFunctor(argums[0].VObject).Functor = TFunctor(o).Functor)and
        ((TFunctor(o).Params <> nil) = (TFunctor(argums[0].VObject).Params <> nil)))
        then unres := 0
        else if ((TFunctor(o).Params <> nil)and(TFunctor(argums[0].VObject).Params <> nil)
        and(TFunctor(o).Params.Count <> TFunctor(argums[0].VObject).Params.Count))
        then unres := 0;
    end
    else if (o is TFunctora) then
    begin
      if not((argums[0].VObject is TFunctora)and
        (TFunctora(argums[0].VObject).Functor = TFunctora(o).Functor))
        then unres := 0;
    end
    else if (o is TNumber) then
    begin
      if not((argums[0].VObject is TNumber)
      and(TNumber(argums[0].VObject).GetNumber = TNumber(o).GetNumber)) then unres := 0
    end
    else if (o is TVariable) then
    begin
      if (argums[0].VObject is TVariable)and(TVariable(o).identifier = TVariable(argums[0].VObject).identifier)
        then begin unres := 1; Exit; end;
      begin
        if  (TId(argums[0].VObject).OccurCheck(TVariable(o)))and
          (
          not(CurProg.Quant1)or(IsHigherObject(o, argums[0].VObject)))
            then
        begin
          mgu1.Add(o);
          mgu1.Add(argums[0].VObject);
          unres := 1;
        end
        else begin unres1 := 0; end;
      end
    end;
  end;
end;

{ Searches for occurence of a variable in an id. }
function TVariable.CompInMgu;
var i : integer;
begin
  Result := self;

  i := 0;
  while (Result = self)and(i < mgu1.Count) do
  begin
    if self = mgu1.Items[i] then Result := mgu1.Items[i+1]
    else
    i := i+2;
  end;
end;


function CompId(o : TObject; var tp : char; argums : array of const) : TObject;
begin
  if o = FiV then ocres := 0;
end;

function TId.OccurCheck;
begin
  ocres := 1;
  FiV := X;
  Dive('6', CompId, [nil]);
  Result := boolean(ocres);
end;

function TId.UnifyId;
var varn : integer;
begin
  varn := mgu1.Count-1;
  repeat
    unres1 := 1;  { Tries to unify id terms until, it is clear that }
    unres := 1;   { no more changes are done.                       }
    varn := mgu1.count;
    Dive('5', IdUnif, [X]);
    if unres1 = 0 then unres := 0;
  until  (varn = mgu1.count)or(unres <> 0);
end;

{ Checks for occurence of a variable. }
procedure CheckForOcc;
var i : integer; X : TVariable;
begin
  i := 0;
  while (i < mgu1.Count)and(unres <> 0) do
  begin
    X := TVariable(mgu1.Items[i]);
    ocres := 1;
    TId(mgu1.Items[i+1]).OccurCheck(X);
    unres := ocres;
    i := i+2;
  end;
end;

var mgucnt : integer;

function TAtom.UnifyS;
var i : integer;
begin
  inc(numsim);
  mgucnt := mgu1.count;
  if (X.Ev <> 0) then begin result := 0; exit; end;
  unres := 1;
  if Id <> nil then Id.UnifyId(X.Id);
  if (unres = 0)and(mgu1.count > mgucnt) then
     for i := mgu1.count-1 downto mgucnt do mgu1.delete(i);
  result := unres;
end;


end.
